from jolt import attributes, Parameter
from jolt.plugins import autotools, cxxinfo, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@autotools.requires()
@cxxinfo.publish(libraries=["tirpc"])
class Libtirpc(autotools.Autotools):
    """Libtirpc package"""
    name = "libtirpc"

    version = Parameter("1.3.7", help="libtirpc version.")
    requires_src = ["fetch:alias=src,url=https://downloads.sourceforge.net/project/libtirpc/libtirpc/1.3.7/libtirpc-1.3.7.tar.bz2"]
    srcdir = "{fetch[src]}/libtirpc-{version}"
    options = ["--disable-gssapi"]


TaskRegistry.get().add_task_class(Libtirpc)



from contextlib import contextmanager
import fasteners
import json
import glob
from importlib.machinery import SourceFileLoader
import os
import platform
import subprocess
import sys
from types import ModuleType

from jolt import inspection
from jolt.tasks import Task, TaskGenerator
from jolt.error import raise_error_if
from jolt import common_pb2 as common_pb
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.tools import Tools


class Recipe(object):
    """
    Abstract representation a single recipe file.

    Implementations of this class are responsible for reading the recipe source
    from the filesystem. The recipe source is then parsed and the tasks are
    extracted and made available for execution.

    The format of the recipe source is implementation defined.
    """

    tasks = []
    """
    List of task classes defined in the recipe.

    Available after the recipe has been loaded.
    """

    def __init__(self, path, joltdir=None, project=None, source=None):
        self.path = path
        self.basepath = os.path.basename(path)
        self.joltdir = joltdir
        self.project = project
        self.source = source
        self.tasks = []

    def load(self):
        """ Load the recipe source from the file system. """
        raise_error_if(self.source is not None, "recipe already loaded: {}", self.path)

        with open(self.path) as f:
            self.source = f.read()

    def save(self):
        """ Save the recipe source to the file system. """
        raise_error_if(self.source is None, "recipe source unknown: {}", self.path)

        with open(self.path, "w") as f:
            f.write(self.source)


class NativeRecipe(Recipe):
    """ Represents a Python recipe file (.jolt, .py). """

    @staticmethod
    def _is_abstract(cls):
        return cls.__dict__.get("abstract", False) or cls.__name__.startswith("_")

    @staticmethod
    def _is_task(cls):
        return isinstance(cls, type) and \
            issubclass(cls, Task) and \
            not NativeRecipe._is_abstract(cls)

    def load(self, joltdir=None):
        """
        Load the recipe source from the file system.

        Python classes defined in the recipe source are extracted and made available
        as tasks for execution. Task classes must be subclasses of Task or TaskGenerator.

        """
        super(NativeRecipe, self).load()

        name = utils.canonical(self.path)
        loader = SourceFileLoader("joltfile_{0}".format(name), self.path)
        module = ModuleType(loader.name)
        module.__file__ = self.path
        loader.exec_module(module)
        sys.modules[loader.name] = module

        classes = inspection.getmoduleclasses(module, [Task, TaskGenerator], NativeRecipe._is_abstract)
        generators = []

        for cls in classes[TaskGenerator]:
            cls.joltdir = os.path.normpath(joltdir or self.joltdir or os.path.dirname(self.path))
            generators.append(cls())

        for generator in generators:
            generated_tasks = utils.as_list(generator.generate())
            classes[Task] += filter(NativeRecipe._is_task, generated_tasks)

        for task in classes[Task]:
            task.name = task.name or task.__name__.lower()
            task.joltdir = os.path.normpath(joltdir or self.joltdir or os.path.dirname(self.path))
            task.joltproject = self.project
            self.tasks.append(task)

        log.verbose("Loaded: {0}", self.path)


class Loader(object):
    """
    Base class for recipe loaders.

    A Loader is responsible for finding recipes in the file system providing a list
    of Recipe:s from which tasks can be loaded.
    """

    def recipes(self) -> list:
        """ Return a list of Recipe:s from which tasks can be loaded. """
        pass


class LoaderFactory(object):
    """
    A factory for creating Loader instances.

    Factories are registered with the JoltLoader where it is used to create Loader instances.
    """
    def create(self):
        raise NotImplementedError()


class NativeLoader(Loader):
    """ A loader for Python recipe files (.jolt, .py). """

    def __init__(self, searchpath):
        """
        Create a new NativeLoader instance.

        Args:
            searchpath (str): The path to search for recipe files. If the path is a file,
                only that file will be loaded. If the path is a directory, all files with
                the .jolt or .py extension will be loaded.

        """

        self._files = self._find_files(searchpath)
        self._recipes = self._load_files(self._files) if self._files else []

    def _find_files(self, searchpath):
        # If the searchpath is a file, load it directly
        if fs.path.isdir(searchpath):
            return glob.glob(fs.path.join(searchpath, "*.jolt"))

        _, ext = fs.path.splitext(searchpath)
        raise_error_if(not fs.path.exists(searchpath), "File does not exist: {}", searchpath)
        raise_error_if(ext not in [".build", ".jolt", ".py"], "Invalid file extension: {}", ext)

        return [searchpath]

    def _load_files(self, files):
        recipes = []
        for filepath in files:
            recipe = NativeRecipe(filepath)
            recipes.append(recipe)
        return recipes

    @property
    def recipes(self):
        return self._recipes


_loaders = []


def register(factory):
    """ Register a LoaderFactory with the JoltLoader. """
    raise_error_if(not issubclass(factory, LoaderFactory),
                   "{} is not a LoaderFactory", factory.__name__)
    _loaders.append(factory)


@register
class NativeLoaderFactory(LoaderFactory):
    """ A factory for creating NativeLoader instances. """
    def create(self, searchpath):
        return NativeLoader(searchpath)


@utils.Singleton
class JoltLoader(object):
    """
    The JoltLoader is responsible for loading recipes from the file system.

    The JoltLoader is a singleton and is used to load recipes from the file system.
    The recipes are loaded from the workspace directory and any project directories
    defined in the workspace. The recipes are then made available for execution.

    """

    filename = "*.jolt"

    def __init__(self):
        self._lock = None
        self._recipes = []
        self._tasks = []
        self._path = None
        self._build_path = None
        self._workspace_name = None

    def _get_first_recipe_path(self):
        for recipe in self.recipes:
            return recipe.path
        return None

    def _find_workspace_path(self, searchdir):
        for factory in _loaders:
            loader = factory().create(searchdir)
            if loader.recipes:
                return searchdir

        parentdir = os.path.dirname(searchdir)
        if searchdir == parentdir:
            return os.getcwd()

        return self._find_workspace_path(parentdir)

    def _get_searchpaths(self):
        return [self.workspace_path]

    def load(self, registry=None):
        """
        Load all recipes from the workspace directory.

        Optionally populate the task registry with the tasks found in the recipes.

        Returns:
            List of Task classes found in the recipes.
        """

        if not self.workspace_path:
            self.set_workspace_path(self._find_workspace_path(os.getcwd()))

        if not self.workspace_path:
            return []

        for searchpath in self._get_searchpaths():
            for factory in _loaders:
                loader = factory().create(searchpath)
                for recipe in loader.recipes:
                    recipe.workspace_path = os.path.relpath(recipe.path, self.workspace_path)
                    recipe.load()
                    self._recipes.append(recipe)
                    self._tasks += recipe.tasks

        # Create workspace lock on the first loaded recipe
        if not self._lock:
            path = self._get_first_recipe_path()
            if path:
                self._lock = fasteners.InterProcessLock(path)

        # Add tasks to the registry if provided
        if registry is not None:
            for task in self._tasks:
                registry.add_task_class(task)

        return self._tasks

    def load_file(self, path, joltdir=None):
        """ Load a single recipe file. """

        for factory in _loaders:
            loader = factory().create(path)
            for recipe in loader.recipes:
                joltdir = fs.path.join(self.joltdir, joltdir) if joltdir else self.joltdir
                recipe.load(joltdir=joltdir)
                self._recipes.append(recipe)
                self._tasks += recipe.tasks

    def load_plugin(self, filepath):
        """ Load a single plugin file. """

        plugin, ext = os.path.splitext(fs.path.basename(filepath))
        loader = SourceFileLoader("jolt.plugins." + plugin, filepath)
        module = ModuleType(loader.name)
        module.__file__ = filepath
        loader.exec_module(module)
        sys.modules[loader.name] = module

    def load_plugins(self):
        """
        Load all configured plugins.

        Plugins are loaded from the plugin path configured in the Jolt configuration file
        or from the default plugin path in the Jolt package.

        If a plugin is already loaded, it will not be loaded again. If a plugin is not found
        in the plugin path, it will not be loaded.
        """
        searchpath = config.get("jolt", "pluginpath")
        searchpath = searchpath.split(":") if searchpath else []
        searchpath.append(fs.path.join(fs.path.dirname(__file__), "plugins"))

        for plugin in config.plugins():
            for path in searchpath:
                if "jolt.plugins." + plugin not in sys.modules:
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin + ".py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin, "__init__.py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue

    @property
    def projects(self):
        return self._recipes

    @property
    def recipes(self):
        return self._recipes

    @property
    def tasks(self):
        """ Returns a list of all loaded tasks. """
        return self._tasks

    @property
    def joltdir(self):
        """ Returns the path to the workspace. """
        return self._path

    @property
    def workspace_name(self):
        """ Returns the name of the workspace. """
        return self._workspace_name or os.path.basename(self.workspace_path)

    def set_workspace_name(self, name):
        self._workspace_name = name

    @property
    def workspace_path(self):
        """ Returns the path to the workspace. """
        return self._path

    @contextmanager
    @utils.locked
    def workspace_lock(self):
        if not self._lock:
            yield
        with self._lock:
            yield

    def set_workspace_path(self, path):
        if not self._path or len(path) < len(self._path):
            self._path = os.path.normpath(path) if path is not None else None

    @property
    def build_path(self):
        """ Returns the path to the build directory. """
        return self._build_path or os.path.join(self.workspace_path, "build")

    @property
    def build_path_rel(self):
        """" Returns the path to the build directory relative to the workspace. """
        return os.path.relpath(self.build_path, self.workspace_path)

    def set_build_path(self, path):
        self._build_path = os.path.normpath(os.path.join(self.workspace_path, path))
        log.debug("Jolt build path: {}", self._build_path)


def get_workspacedir():
    workspacedir = JoltLoader.get().workspace_path
    assert workspacedir is not None, "No workspace present"
    return workspacedir


@contextmanager
def workspace_lock():
    with JoltLoader.get().workspace_lock():
        yield


def workspace_locked(func):
    def wrapper(*args, **kwargs):
        with workspace_lock():
            return func(*args, **kwargs)
    return wrapper


def import_workspace(buildenv: common_pb.BuildEnvironment):
    """
    Import workspace from a BuildEnvironment protobuf message.

    This function will create files, recipes, resources and modules in the workspace
    based on the information in the BuildEnvironment message.

    The workspace tree is not pulled and checked out here. This is done by the
    worker before it starts the executor.

    """
    loader = JoltLoader.get()
    loader.set_workspace_path(os.getcwd())
    loader.set_workspace_name(buildenv.workspace.name)
    if buildenv.workspace.builddir:
        loader.set_build_path(buildenv.workspace.builddir)

    # Write .jolt files into workspace
    for file in buildenv.workspace.files:
        if file.content:
            with open(file.path, "w") as f:
                f.write(file.content)


@workspace_locked
def export_workspace(tasks=None) -> common_pb.Workspace:
    """
    Export workspace to a Workspace protobuf message.

    This function will create a Workspace protobuf message containing all the
    recipes, resources and modules in the workspace. If tasks is provided, only
    the projects associated with the tasks will be exported. Otherwise, all
    projects will be exported.

    If the workspace is configured to use a remote cache, the workspace will be
    pushed to the remote cache using the fstree tool. The tree hash of the
    workspace will be included in the returned Workspace message.

    """
    loader = JoltLoader.get()
    tools = Tools()
    tree = None

    fstree_enabled = config.getboolean("jolt", "fstree", True)
    if fstree_enabled:
        fstree = tools.which("fstree")
        if not fstree:
            host = platform.system().lower()
            arch = platform.machine().lower()
            fstree = os.path.join(os.path.dirname(__file__), "bin", f"fstree-{host}-{arch}")
        if not os.path.exists(fstree):
            fstree_enabled = False
            log.warning("fstree executable not found, will not push workspace to remote cache")

    cache_grpc_uri = config.geturi("cache", "grpc_uri")
    if fstree_enabled:
        if not cache_grpc_uri:
            fstree_enabled = False
            log.warning("No cache gRPC URI configured, will not push workspace to remote cache")
        else:
            raise_error_if(cache_grpc_uri.scheme not in ["tcp"], "Invalid scheme in cache gRPC URI config: {}", cache_grpc_uri.scheme)
            raise_error_if(not cache_grpc_uri.netloc, "Invalid network address in cache gRPC URI config: {}", cache_grpc_uri.netloc)

    # Push workspace to remote cache if possible
    if fstree_enabled and fstree and cache_grpc_uri:
        with tools.cwd(loader.workspace_path):
            cwd = tools.getcwd()
            cachedir = config.get_cachedir()
            indexhash = utils.hashstring(cwd)
            indexfile = tools.expand_path(
                "{cachedir}/indexes/{}/{}",
                indexhash[:2], indexhash[2:],
                cwd=fs.posixpath.abspath(cwd),
                cachedir=cachedir)

            if not os.path.exists(indexfile):
                process = None
                try:
                    with log.progress("Indexing workspace for the first time", count=None, unit="objects", estimates=False) as progress:
                        process = subprocess.Popen(
                            [fstree, "write-tree", "--json", "--cache", cachedir, "--ignore", ".joltignore", "--index", indexfile, "--remote", cache_grpc_uri.geturl(), "--threads", str(tools.thread_count())],
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.PIPE,
                            cwd=tools.getcwd())

                        for line in iter(process.stderr.readline, b''):
                            try:
                                event = json.loads(line.decode())
                            except json.JSONDecodeError as exc:
                                log.error("Failed to decode fstree event: {}", line.decode())
                                raise exc

                            if event.get("type") in ["cache::add"]:
                                progress.update(1)

                except Exception as exc:
                    if process:
                        process.terminate()
                    raise exc
                finally:
                    if process:
                        process.wait()
                        process.stderr.close()
                        raise_error_if(process.returncode != 0, "Failed to index workspace")

            process = None
            try:
                with log.progress("Pushing workspace to remote cache", count=None, unit="objects", estimates=False) as progress:
                    process = subprocess.Popen(
                        [fstree, "write-tree-push", "--json", "--cache", cachedir, "--ignore", ".joltignore", "--index", indexfile, "--remote", cache_grpc_uri.geturl(), "--threads", str(tools.thread_count())],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.PIPE,
                        cwd=tools.getcwd())

                    for line in iter(process.stderr.readline, b''):
                        try:
                            event = json.loads(line.decode())
                        except json.JSONDecodeError as exc:
                            log.error("Failed to decode fstree event: {}", line.decode())
                            raise exc

                        if event.get("type") in ["cache::push"]:
                            tree = event.get("path")
                            log.info("Workspace tree: {} ({} objects)", tree, event.get("value"))

                        if event.get("type") in ["cache::remote_missing_object", "cache::remote_missing_tree"]:
                            progress.refresh()

                        if event.get("type") in ["cache::push_object", "cache::push_tree"]:
                            progress.update(1)

            except Exception as exc:
                if process:
                    process.terminate()
                raise exc

            finally:
                if process:
                    process.wait()
                    process.stderr.close()
                    raise_error_if(process.returncode != 0, "Failed to push workspace to remote cache")

    workspace = common_pb.Workspace(
        builddir=loader.build_path_rel,
        cachedir=config.get_cachedir(),
        rootdir=loader.workspace_path,
        name=loader.workspace_name,
        tree=tree,
    )

    if not fstree_enabled:
        for recipe in loader.recipes:
            workspace.files.append(
                common_pb.File(
                    path=recipe.workspace_path,
                    content=recipe.source,
                )
            )

    return workspace



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class RapidJSON(cmake.CMake):
    name = "rapidjson"
    version = Parameter("24b5e7a", help="rapidjson version.")
    options = ["CMAKE_POLICY_VERSION_MINIMUM=3.5"]
    requires_git = ["git:url=https://github.com/Tencent/rapidjson.git,rev={version}"]
    srcdir = "{git[rapidjson]}"
    options = [
        "RAPIDJSON_BUILD_EXAMPLES=OFF",
        "RAPIDJSON_BUILD_TESTS=OFF",
    ]


TaskRegistry.get().add_task_class(RapidJSON)



import functools
from jolt import attributes, include
from jolt import influence
from jolt import Task
from jolt import utils
from jolt.error import raise_task_error_if

import os


def options(attrib):
    """
    Decorates a CMake task with an alternative ``options`` attribute.

    The new attribute will be concatenated with the regular
    ``options`` attribute.

    Args:
        attrib (str): Name of alternative attribute.
            Keywords are expanded.
    """
    return utils.concat_attributes("options", attrib)


def requires(version=None):
    """ Decorator to add CMake requirements to a task. """

    import jolt.pkgs.cmake

    def decorate(cls):
        cls = attributes.requires("requires_cmake")(cls)
        cls.requires_cmake = ["cmake" + (f":version={version}" if version else "")]
        return cls

    return decorate


def use_ninja():
    """
    Decorator to add Ninja dependencies to CMake task.

    It also selects Ninja as the CMake generator and builder.
    """

    import jolt.pkgs.ninja

    def decorate(cls):
        cls = attributes.requires("requires_ninja")(cls)
        cls.generator = "Ninja"
        cls.requires_ninja = ["ninja"]
        return cls

    return decorate


@attributes.common_metadata()
@options("options")
class CMake(Task):
    """ Builds and publishes a CMake project """
    abstract = True

    cmakelists = "CMakeLists.txt"
    """ Path to CMakeLists.txt or directory containing CMakelists.txt """

    config = "Release"
    """ The default build configuration to use """

    generator = None
    """ The build file generator that CMake should use """

    incremental = True
    """ Keep build directories between Jolt invocations """

    options = []
    """ List of options and their values (``option[:type]=value``) """

    srcdir = None
    """ Source directory. If not specified, the task working directory is used. """

    install = ["install"]
    """ List of install build targets. If empty, the default install target is used. """

    def clean(self, tools):
        cmake = tools.cmake(incremental=self.incremental)
        cmake.clean()

    def run(self, deps, tools):
        self.deps = deps
        raise_task_error_if(not self.cmakelists, self, "cmakelists attribute has not been defined")

        config = tools.expand(str(self.config))
        options = self._options()
        options += ["CMAKE_BUILD_TYPE=" + config]

        with tools.cwd(self.srcdir or self.joltdir):
            cmake = tools.cmake(deps, incremental=self.incremental)
            cmake.configure(tools.expand(self.cmakelists), *["-D" + tools.expand(option) for option in options], generator=self.generator)
            for target in self.install or ["install"]:
                cmake.install(config=config, target=target)

    def publish(self, artifact, tools):
        cmake = tools.cmake(incremental=self.incremental)
        cmake.publish(artifact)


@influence.attribute("headers", type=influence.FileInfluence)
@influence.attribute("sources", type=influence.FileInfluence)
class _CMakeCXX(CMake):
    asflags = []
    binary = None
    cflags = []
    cxxflags = []
    headers = []
    incpaths = []
    ldflags = []
    macros = []
    shared = False
    sources = []

    generator = None
    incremental = True
    template = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.binary or self.canonical_name

    def run(self, deps, tools):
        with tools.cwd(tools.builddir(incremental=self.incremental)):
            self.cmakelists = tools.expand_path(self.cmakelists)
            project = utils.render(self.template, deps=deps, task=self, tools=tools, os=os)
            tools.write_file(self.cmakelists, project, expand=False)
        super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.headers:
            for header in self.headers:
                artifact.collect(header)
            artifact.cxxinfo.incpaths.append(".")


class CXXLibrary(_CMakeCXX):
    abstract = True
    template = "cxxlibrary.cmake.template"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.headers:
            for header in self.headers:
                artifact.collect(header)
            artifact.cxxinfo.incpaths.append(".")
        artifact.cxxinfo.libraries.append("{binary}")
        artifact.cxxinfo.libpaths.append("lib")


class CXXExecutable(_CMakeCXX):
    abstract = True
    template = "cxxexecutable.cmake.template"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)



from jolt import attributes, Parameter
from jolt.plugins import cmake, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@cmake.requires()
@cmake.use_ninja()
class Patchelf(cmake.CMake):
    name = "patchelf"
    version = Parameter("0.15.5", help="patchelf version.")
    requires_src = ["fetch:alias=src,url=https://github.com/NixOS/patchelf/archive/refs/tags/{version}.tar.gz"]
    srcdir = "{fetch[src]}/patchelf-{version}"


TaskRegistry.get().add_task_class(Patchelf)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.tasks import TaskRegistry
from jolt.plugins import git, cmake


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class GoogleBenchmark(cmake.CMake):
    name = "google/benchmark"
    version = Parameter("1.9.4", help="Benchmark version.")
    requires_git = ["git:url=https://github.com/google/benchmark.git,rev=v{version},submodules=true"]
    options = ["BENCHMARK_ENABLE_TESTING=OFF"]
    srcdir = "{git[benchmark]}"


TaskRegistry.get().add_task_class(GoogleBenchmark)



from jolt import attributes, Parameter, Task
from jolt.plugins import git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
class Cista(Task):
    name = "cista"
    version = Parameter("0.16", help="Cista version.")
    requires_git = ["git:url=https://github.com/felixguendling/cista.git,rev=v{version}"]
    srcdir = "{git[cista]}"

    def publish(self, artifact, tools):
        with tools.cwd(self.srcdir):
            artifact.collect("include/")
        artifact.cxxinfo.incpaths.append("include")


TaskRegistry.get().add_task_class(Cista)



from jolt import Download, Parameter, attributes
from jolt.pkgs import rust
from jolt.plugins import git, rust
from jolt.tasks import TaskRegistry


@attributes.attribute("triple", "triple_{system}_{arch}")
@attributes.arch
@attributes.system
class Rust(Download):
    name = "rust"
    version = Parameter("1.92.0", help="Rust version.")
    url = "https://static.rust-lang.org/dist/rust-{version}-{triple}.tar.xz"
    collect = [{"files": "*", "cwd": "rust-{version}-{triple}"}]
    components = ["cargo", "rustc", "rust-std-{triple}"]

    triple_linux_x86_64 = "x86_64-unknown-linux-gnu"
    triple_windows_x86_64 = "x86_64-pc-windows-msvc"
    triple_darwin_x86_64 = "x86_64-apple-darwin"

    def publish(self, artifact, tools):
        self.installdir = tools.builddir("install")
        with tools.cwd(self._extractdir, "rust-{version}-{triple}"):
            for component in tools.expand(self.components):
                artifact.collect("*", cwd=component)
        artifact.environ.PATH.append("bin")


@attributes.requires("requires_git")
@attributes.requires("requires_rust")
class RustBindgen(rust.Rust):
    name = "rust-bindgen"
    version = Parameter("0.72.1", help="Rust Bindgen version.")

    requires_git = ["git:url=https://github.com/rust-lang/rust-bindgen.git,rev=v{version}"]
    requires_rust = ["rust"]
    srcdir = "{git[rust-bindgen]}/bindgen-cli"


TaskRegistry.get().add_task_class(Rust)
TaskRegistry.get().add_task_class(RustBindgen)



from jolt import attributes, influence
from jolt import Alias, Download, Parameter, Task
from jolt.tasks import TaskRegistry
from jolt.error import raise_task_error_if


class StrawberryPerl(Download):
    name = "perl/strawberry"
    version = Parameter("5.42.0.1", help="poco version.")
    url = "https://github.com/StrawberryPerl/Perl-Dist-Strawberry/releases/download/SP_{version_compact}_64bit/strawberry-perl-{version}-64bit-portable.zip"

    @property
    def version_compact(self):
        return str(self.version).replace(".", "")

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append("perl/bin")


@attributes.system
@influence.string("{system}")
class HostPerl(Task):
    name = "perl/host"

    def run(self, deps, tools):
        raise_task_error_if(not tools.which("perl"), "Perl is not installed on the host system.")


@attributes.system
@attributes.requires("requires_{system}")
class Perl(Alias):
    name = "virtual/perl"

    requires_darwin = ["perl/host"]
    requires_linux = ["perl/host"]
    requires_windows = ["perl/strawberry"]


TaskRegistry.get().add_task_class(Perl)
TaskRegistry.get().add_task_class(HostPerl)
TaskRegistry.get().add_task_class(StrawberryPerl)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["double-conversion"])
class DoubleConversion(cmake.CMake):
    name = "double-conversion"
    version = Parameter("3.4.0", help="double_conversion version.")
    options = ["BUILD_SHARED_LIBS=ON"]
    requires_git = ["git:url=https://github.com/google/double-conversion.git,rev=v{version}"]
    srcdir = "{git[double-conversion]}"


TaskRegistry.get().add_task_class(DoubleConversion)



import os
import sys

from jolt import Task
from jolt import attributes
from jolt import filesystem as fs
from jolt.cache import ArtifactStringAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider
from jolt.error import raise_task_error_if


class PythonVariable(ArtifactStringAttribute):
    def __init__(self, artifact, name):
        super(PythonVariable, self).__init__(artifact, name)

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass


class PythonListVariable(PythonVariable):
    def __init__(self, artifact, name):
        super(PythonListVariable, self).__init__(artifact, name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + ":" + value)
        else:
            self.set_value(value)

    def items(self):
        value = self.get_value()
        return value.split(":") if value is not None else []


class PythonPathVariable(PythonListVariable):
    def __init__(self, artifact, name):
        super(PythonPathVariable, self).__init__(artifact, name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + ":" + value)
        else:
            self.set_value(value)

    def items(self):
        value = self.get_value()
        return value.split(":") if value is not None else []

    def apply(self, task, artifact):
        paths = [fs.path.join(artifact.path, path) for path in self.items()]
        sys.path.extend(paths)

    def unapply(self, task, artifact):
        paths = [fs.path.join(artifact.path, path) for path in self.items()]
        for path in paths:
            sys.path.remove(path)


class PythonDictVariable(PythonListVariable):
    def __init__(self, artifact, name):
        super(PythonDictVariable, self).__init__(artifact, name)

    def __setitem__(self, key, value=None):
        item = "{0}={1}".format(key, value) if value is not None else key
        self.append(item)


class Python(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(Python, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name.lower() == "path":
            return PythonPathVariable(self._artifact, "PATH")
        assert False, "No such python attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class PythonProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "python", Python(artifact))

    def parse(self, artifact, content):
        if "python" not in content:
            return
        for key, value in content["python"].items():
            setattr(artifact.python, key, value)

    def format(self, artifact, content):
        if "python" not in content:
            content["python"] = {}
        for key, value in artifact.python.items():
            content["python"][key] = str(value)

    def apply(self, task, artifact):
        artifact.python.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.python.unapply(task, artifact)


@attributes.system
@attributes.requires("requires_python")
class PythonEnv(Task):
    """
    Base class for Python virtual environment tasks.

    Builds a Python virtual environment and installs specified packages.

    The venv module from the Python standard library must be available in the
    Python installation used to run the task.
    """

    abstract = True
    """ This is an abstract base class that should be inherited by concrete tasks. """

    executable = "python3"
    """ Python executable to use for creating the virtual environment. """

    requirements = []
    """
    List of Python packages to install in the virtual environment.

    Each entry should be a string suitable for pip, e.g., "package==version".
    """

    def _verify_influence(self, deps, artifact, tools, sources=None):
        # No influence to verify
        return

    def relocate_scripts(self, artifact, tools, frompath, topath):
        bindir = "Scripts" if self.system == "windows" else "bin"

        with tools.cwd(artifact.path, bindir):
            for script in tools.glob("*"):
                if script.startswith("python"):
                    continue
                if os.path.isdir(tools.expand_path(script)):
                    continue
                tools.replace_in_file(script, frompath, topath)

        with tools.cwd(artifact.path):
            if not tools.exists("local/bin"):
                return
            with tools.cwd("local", "bin"):
                for script in tools.glob("*"):
                    tools.replace_in_file(script, frompath, topath)

    def publish(self, artifact, tools):
        # Create a parallel installation by copying a Python installation

        # First locate the Python executable to copy
        py_exe = tools.which(self.executable)
        raise_task_error_if(
            py_exe is None, self,
            f"Python executable '{self.executable}' not found in PATH.",
        )

        # Follow symlinks to get the real executable
        py_exe = fs.path.realpath(py_exe)

        # Determine the Python home directory
        py_home = fs.path.dirname(fs.path.dirname(py_exe))

        # Determine the Python version
        self.version_major = tools.run(
            [py_exe, "-c", "import sys; print(\"{{}}.{{}}\".format(sys.version_info[0], sys.version_info[1]))"],
            shell=False,
            output_on_error=True).strip()

        self.info("Python executable: {0}", py_exe)
        self.info("Python home: {0}", py_home)
        self.info("Python version: {0}", self.version_major)

        # Copy the Python installation to the artifact path
        with tools.cwd(py_home):
            artifact.collect(py_exe, "bin/python3")
            artifact.collect("lib/python3")
            artifact.collect("lib/python{version_major}")
            artifact.collect("lib/libpython{version_major}.*")

        # Create common symlinks
        if self.system != "windows":
            with tools.cwd(artifact.path, "bin"):
                tools.symlink("python3", "python")
                tools.symlink("python3", "python{version_major}")
            with tools.cwd(artifact.path, "lib", "python{version_major}"):
                tools.unlink("sitecustomize.py", ignore_errors=True)

        # Install required packages into the artifact using pip
        with tools.environ(PYTHONHOME=artifact.path):
            py_exe = fs.path.join(artifact.path, "bin", "python3")
            with tools.tmpdir() as tmp, tools.cwd(tmp):
                tools.write_file(
                    "requirements.txt",
                    "\n".join(self.requirements) + "\n"
                )

                pip_cmd = [
                    py_exe,
                    "-m",
                    "pip",
                    "--isolated",
                    "--no-cache-dir",
                    "install",
                    "-r",
                    "requirements.txt",
                    "--prefix", artifact.path,
                    "--break-system-packages",
                ]
                tools.run(pip_cmd, shell=False)

        artifact.environ.PATH.append("bin")
        artifact.environ.PATH.append("local/bin")
        artifact.strings.install_prefix = artifact.path

    def unpack(self, artifact, tools):
        # Relocate the virtual environment by adjusting script paths
        frompath = artifact.strings.install_prefix
        topath = artifact.final_path
        self.relocate_scripts(artifact, tools, frompath, topath)

        artifact.strings.install_prefix = artifact.final_path


def requires(python=True):
    """ Decorator to add Python requirements to a task. """

    import jolt.pkgs.cpython

    def decorate(cls):
        if python:
            cls = attributes.requires("requires_python")(cls)
            cls.requires_python = ["cpython"]

        return cls

    return decorate



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake, protobuf
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class OpenCV(cmake.CMake):
    name = "opencv"
    version = Parameter("4.12.0", help="OpenCV version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/opencv/opencv.git,rev={version}"]
    srcdir = "{git[opencv]}"
    options = [
        "BUILD_EXAMPLES=OFF",
        "BUILD_TESTS=OFF",
        "BUILD_PERF_TESTS=OFF",
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(OpenCV)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.pkgs import nasm, perl
from jolt.plugins import cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_perl")
@attributes.requires("requires_{system}_nasm")
@attributes.common_metadata()
@attributes.system
@cxxinfo.publish(libraries=["ssl", "crypto"])
class OpenSSL(Task):
    name = "openssl"
    version = Parameter("3.6.0", help="openssl version.")
    shared = BooleanParameter(False, "Enable shared libraries.")

    requires_git = ["git:url=https://github.com/openssl/openssl.git,rev=openssl-{version}"]
    requires_perl = ["virtual/perl"]
    requires_windows_nasm = ["nasm"]

    def clean(self):
        pass

    def run(self, deps, tools):
        self.builddir = tools.builddir(incremental=True)
        self.installdir = tools.builddir("install")
        self.srcdir = tools.expand_path("{git[openssl]}")
        with tools.cwd(self.builddir):
            self.info("Configuring OpenSSL... {builddir}")

            if self.system == "windows":
                tools.run("perl {srcdir}/Configure --prefix={installdir} --openssldir=ssl no-tests {shared[,no-shared]}")
                tools.run("nmake")
                tools.run("nmake install")
            else:
                tools.run("{srcdir}/config --prefix={installdir} no-tests {shared[,no-shared]}")
                tools.run("make -j{}", tools.cpu_count())
                tools.run("make install")

    def publish(self, artifact, tools):
        with tools.cwd(self.installdir):
            artifact.collect("*", symlinks=True)
        artifact.strings.install_prefix = self.installdir
        if self.system == "windows":
            self.publish_pkgconfig(artifact, tools)

    def publish_pkgconfig(self, artifact, tools):
        with tools.cwd(artifact.path):
            for libdir in ["lib", "lib32", "lib64"]:
                if tools.exists(libdir):
                    self.libdir = libdir
                    break

        with tools.tmpdir() as tmp, tools.cwd(tmp):
            tools.write_file(
                "openssl.pc",
                """
# See: man pkg-config
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/{libdir}
includedir=${{prefix}}/include

Name: OpenSSL
Description: Secure Sockets Layer and cryptography libraries and tools
Version: {version}-{identity}
Requires: libssl libcrypto
""")
            tools.write_file(
                "libcrypto.pc",
                """
# See: man pkg-config
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/{libdir}
includedir=${{prefix}}/include
enginesdir=${{libdir}}/engines-3

Name: OpenSSL-libcrypto
Description: OpenSSL cryptography library
Version: {version}-{identity}
Libs: -L${{libdir}} -lcrypto
Cflags: -I${{includedir}}
""")
            tools.write_file(
                "libssl.pc",
                """
# See: man pkg-config
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/{libdir}
includedir=${{prefix}}/include

Name: OpenSSL-libssl
Description: Secure Sockets Layer and cryptography libraries
Version: {version}-{identity}
Libs: -L${{libdir}} -lssl
Cflags: -I${{includedir}}
""")
            artifact.collect("openssl.pc", "{libdir}/pkgconfig/")
            artifact.collect("libcrypto.pc", "{libdir}/pkgconfig/")
            artifact.collect("libssl.pc", "{libdir}/pkgconfig/")


TaskRegistry.get().add_task_class(OpenSSL)



from jolt import Alias, BooleanParameter, Download, ListParameter, Parameter, Task, attributes
from jolt.pkgs import cmake, ninja
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


VERSION = "21.1.8"


@attributes.common_metadata()
class LLVMFromBin(Download):
    name = "llvm/bin"
    version = Parameter("21.1.0", help="LLVM version.")
    url = [
        "https://github.com/llvm/llvm-project/releases/download/llvmorg-{version}/LLVM-{version}-Linux-X64.tar.xz"
    ]

    collect = [{"files": "*", "cwd": "LLVM-{version}-Linux-X64"}]


@attributes.requires("requires_git")
@attributes.requires("requires_ninja")
@cmake.requires()
@cmake.use_ninja()
class LLVMFromSrc(cmake.CMake):
    name = "llvm/src"

    projects = ListParameter(
        ["clang", "clang-tools-extra", "lld", "libclc"],
        values=[
            "clang",
            "clang-tools-extra",
            "lld",
            "libclc",
            "libcxx",
            "libcxxabi",
            "compiler-rt",
            "openmp",
            "polly",
        ],
        help="LLVM projects to build.",
    )
    release = BooleanParameter(True, help="Build a release version.")
    rtti = BooleanParameter(True, help="Enable RTTI support.")
    targets = ListParameter(
        ["X86", "ARM", "AArch64", "RISCV", "AMDGPU"],
        values=["X86", "AMDGPU", "ARM", "AArch64", "RISCV", "Mips", "PowerPC", "SystemZ"],
        help="LLVM targets to build.",
    )
    version = Parameter(VERSION, help="LLVM version.")

    requires_ninja = ["ninja"]
    requires_git = ["git:url=https://github.com/llvm/llvm-project.git,rev=llvmorg-{version}"]

    options = [
        "CMAKE_BUILD_TYPE={release[Release,Debug]}",
        "LLVM_ENABLE_PROJECTS='{_projects_str}'",
        "LLVM_TARGETS_TO_BUILD='{_targets_str}'",
        "LLVM_ENABLE_RTTI={rtti[ON,OFF]}",
    ]

    srcdir = "{git[llvm-project]}/llvm"

    @property
    def _projects_str(self):
        return ";".join(self.projects)

    @property
    def _targets_str(self):
        return ";".join(self.targets)


class LLVM(Alias):
    name = "llvm"
    version = Parameter(VERSION, help="LLVM version.")
    requires = ["llvm/src:version={version},release=true"]


TaskRegistry.get().add_task_class(LLVMFromBin)
TaskRegistry.get().add_task_class(LLVMFromSrc)
TaskRegistry.get().add_task_class(LLVM)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires(version="3.31.1")
class Libyajl(cmake.CMake):
    name = "libyajl"
    version = Parameter("2.1.0", help="libyajl version.")
    requires_git = ["git:url=https://github.com/lloyd/yajl.git,rev={version}"]
    srcdir = "{git[yajl]}"


TaskRegistry.get().add_task_class(Libyajl)



from jolt import attributes, Parameter
from jolt.pkgs import x11
from jolt.tasks import TaskRegistry
from jolt.plugins import meson, git, pkgconfig


@attributes.arch
@attributes.requires("requires_git")
@attributes.requires("requires_glproto")
@attributes.requires("requires_x11")
@attributes.requires("requires_xext")
@meson.requires()
@pkgconfig.requires()
class Libglvnd(meson.Meson):
    name = "libglvnd"
    version = Parameter("1.7.0", help="libglvnd version.")

    requires_git = ["git:url=https://github.com/NVIDIA/libglvnd.git,rev=v{version}"]
    requires_glproto = ["glproto"]
    requires_x11 = ["libx11"]
    requires_xext = ["libxext"]
    srcdir = "{git[libglvnd]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.CMAKE_PREFIX_PATH.append(".")
        artifact.environ.LD_LIBRARY_PATH.append("lib/{arch}-linux-gnu")


TaskRegistry.get().add_task_class(Libglvnd)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import ncurses
from jolt.plugins import cxxinfo, git, autotools, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_ncurses")
@autotools.requires()
@pkgconfig.requires()
@cxxinfo.publish(libraries=["edit"])
class Libedit(autotools.Autotools):
    name = "libedit"
    version = Parameter("20251016-3.1", help="Libedit version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["fetch:alias=src,url=https://thrysoee.dk/editline/libedit-{version}.tar.gz"]
    requires_ncurses = ["ncurses=ncurses:shared=true"]  # No pkgconfig module without shared libs
    srcdir = "{fetch[src]}/libedit-{version}"
    options = [
        "--disable-examples",
        "--{shared[with,without]}-shared",
        "--enable-pc-files",
        "--with-pkg-config-libdir=/jolt-prefix/lib/pkgconfig",
    ]

    def run(self, deps, tools):
        with tools.environ(
            CFLAGS=tools.run("pkg-config --cflags ncurses").strip() + " " + tools.getenv("CFLAGS", ""),
            LDFLAGS=tools.run("pkg-config --libs ncurses").strip() + " " + tools.getenv("LDFLAGS", ""),
        ):
            super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.CMAKE_PREFIX_PATH.append(".")
        if self.shared:
            artifact.environ.LD_LIBRARY_PATH.append("lib")


TaskRegistry.get().add_task_class(Libedit)



from jolt import attributes, Parameter
from jolt.pkgs import libdrm, meson
from jolt.plugins import git, meson
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_libdrm")
@meson.requires()
class Libva(meson.Meson):
    name = "libva"
    version = Parameter("2.16.0", help="Libva version.")
    requires_git = ["git:url=https://github.com/intel/libva.git,rev={version}"]
    requires_libdrm = ["libdrm"]
    srcdir = "{git[libva]}"


TaskRegistry.get().add_task_class(Libva)



from jolt import attributes, BooleanParameter, Parameter
from jolt.tasks import TaskRegistry
from jolt.plugins import cmake, cxxinfo, git
from jolt.pkgs import bzip2, libexpat, lz4, openssl, xz, zlib, zstd


@attributes.requires("requires_bzip2")
@attributes.requires("requires_expat")
@attributes.requires("requires_lz4")
@attributes.requires("requires_openssl")
@attributes.requires("requires_xz")
@attributes.requires("requires_zlib")
@attributes.requires("requires_zstd")
@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["archive"])
class Libarchive(cmake.CMake):
    name = "libarchive"
    version = Parameter("3.8.5", help="libarchive version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_bzip2 = ["libbzip2"]
    requires_git = ["git:url=https://github.com/libarchive/libarchive.git,rev=v{version}"]
    requires_expat = ["libexpat"]
    requires_lz4 = ["lz4"]
    requires_openssl = ["openssl"]
    requires_xz = ["xz"]
    requires_zlib = ["zlib"]
    requires_zstd = ["zstd"]
    srcdir = "{git[libarchive]}"
    options = [
        "CMAKE_POLICY_VERSION_MINIMUM=3.5",
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "ENABLE_TEST=OFF",
    ]


TaskRegistry.get().add_task_class(Libarchive)



from jolt import Alias
from jolt.tasks import TaskRegistry
from jolt.pkgs import python


class Meson(Alias):
    """
    Alias for the Meson build system.

    Meson is included in the Python virtual environment task.
    """

    name = "meson"
    requires = ["python"]


TaskRegistry.get().add_task_class(Meson)



from jolt import attributes, Parameter, Task
from jolt.pkgs import autoconf, automake, perl, zlib
from jolt.plugins import autotools, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_autotools_{system}")
@attributes.requires("requires_git")
@attributes.requires("requires_perl")
@attributes.requires("requires_zlib")
@attributes.system
class NASM(Task):
    name = "nasm"
    version = Parameter("3.01", help="nasm version.")
    requires_autotools_darwin = ["autoconf", "automake"]
    requires_autotools_linux = requires_autotools_darwin
    requires_git = ["git:url=https://github.com/netwide-assembler/nasm.git,rev=nasm-{version}"]
    requires_zlib = ["zlib"]
    requires_perl = ["virtual/perl"]
    srcdir = "{git[nasm]}"

    def run(self, deps, tools):
        self.srcdir = tools.expand_path(self.srcdir)
        self.builddir = tools.builddir()

        if self.system != "windows":
            with tools.cwd(self.srcdir):
                tools.run("{srcdir}/autogen.sh")

            with tools.cwd(self.builddir):
                tools.run("{srcdir}/configure")
                tools.run("make -j{}", tools.thread_count())
        else:
            with tools.cwd(self.srcdir):
                tools.run("nmake -f Mkfiles/msvc.mak")

    def publish(self, artifact, tools):
        if self.system == "windows":
            with tools.cwd(self.srcdir):
                artifact.collect("nasm.exe", "bin/")
                artifact.collect("ndisasm.exe", "bin/")
        else:
            with tools.cwd(self.builddir):
                artifact.collect("nasm", "bin/")
                artifact.collect("ndisasm", "bin/")
        artifact.environ.PATH.append("bin")


TaskRegistry.get().add_task_class(NASM)



from jolt import attributes
from jolt.pkgs import rust
from jolt.plugins import rust
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_rust")
class CBindgen(rust.Rust):
    name = "cbindgen"
    version = "0.29.2"
    requires_git = ["git:url=https://github.com/mozilla/cbindgen.git,rev=v{version},hash=true"]
    requires_rust = ["rust"]
    srcdir = "{git[cbindgen]}"


TaskRegistry.get().add_task_class(CBindgen)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class TomlPlusPlus(cmake.CMake):
    name = "tomlplusplus"
    version = Parameter("3.4.0", help="TomlPlusPlus version.")
    tests = BooleanParameter(False, help="Build tests.")
    requires_git = ["git:url=https://github.com/marzer/tomlplusplus.git,rev=v{version}"]
    srcdir = "{git[tomlplusplus]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.cxxinfo.incpaths.append("include")


TaskRegistry.get().add_task_class(TomlPlusPlus)



from jolt import BooleanParameter, attributes, Alias, Parameter, Task
from jolt.plugins import cxxinfo, git, autotools
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@autotools.requires()
@cxxinfo.publish(libraries=["sqlite3"])
class SQLitePosix(autotools.Autotools):
    name = "sqlite/posix"
    version = Parameter("3.51.1", help="sqlite version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/sqlite/sqlite.git,rev=version-{version}"]
    srcdir = "{git[sqlite]}"
    options = [
        "--{shared[enable,disable]}-shared",
    ]


@attributes.common_metadata()
@attributes.requires("requires_git")
@cxxinfo.publish(libraries=["sqlite3"])
class SQLiteWin32(Task):
    name = "sqlite/win32"
    version = Parameter("3.51.1", help="sqlite version.")
    shared = BooleanParameter(True, help="Build shared libraries.")
    requires_git = ["git:clean=true,url=https://github.com/sqlite/sqlite.git,rev=version-{version}"]
    srcdir = "{git[sqlite]}"

    def run(self, deps, tools):
        self.srcdir = tools.expand_path(self.srcdir)
        with tools.cwd(self.srcdir):
            tools.run("nmake -f Makefile.msc sqlite3.exe sqlite3.dll")
            tools.write_file(
                "sqlite3.pc",
                """
prefix=${{pcfiledir}}/../..
exec_prefix=${{prefix}}/bin
libdir=${{prefix}}/lib
includedir=${{prefix}}/include

Name: SQLite
Description: SQL database engine
Version: {version}-{identity}
Libs: -L${{libdir}} -lsqlite3
Cflags: -I${{includedir}}
""")

    def publish(self, artifact, tools):
        with tools.cwd(self.srcdir):
            if self.shared:
                artifact.collect("sqlite3.dll", "bin/")
                artifact.collect("sqlite3.lib", "lib/")
            else:
                artifact.collect("libsqlite3.lib", "lib/")
            artifact.collect("sqlite3.exe", "bin/")
            artifact.collect("sqlite*.h", "include/")
            artifact.collect("sqlite3.pc", "lib/pkgconfig/")
            artifact.environ.CMAKE_PREFIX_PATH.append(".")


@attributes.requires("requires_{system}")
@attributes.system
class SQLite(Alias):
    name = "sqlite"
    version = Parameter("3.51.1", help="sqlite version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_darwin = ["sqlite/posix:shared={shared},version={version}"]
    requires_linux = requires_darwin
    requires_windows = ["sqlite/win32:shared={shared},version={version}"]


TaskRegistry.get().add_task_class(SQLite)
TaskRegistry.get().add_task_class(SQLitePosix)
TaskRegistry.get().add_task_class(SQLiteWin32)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import openssl
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["zmq"])
class Libzmq(cmake.CMake):
    name = "libzmq"
    version = Parameter("4.3.5", help="ZeroMQ version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/zeromq/libzmq.git,rev=v{version}"]
    srcdir = "{git[libzmq]}"
    options = [
        "BUILD_SHARED={shared[ON,OFF]}",
        "BUILD_TESTS=OFF",
        "CMAKE_POLICY_VERSION_MINIMUM=3.5",
    ]


@attributes.requires("requires_git")
@attributes.requires("requires_libzmq")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Cppzmq(cmake.CMake):
    name = "cppzmq"
    version = Parameter("4.11.0", help="cppzmq version.")
    requires_git = ["git:url=https://github.com/zeromq/cppzmq.git,rev=v{version}"]
    requires_libzmq = ["libzmq"]
    srcdir = "{git[cppzmq]}"
    options = ["CPPZMQ_BUILD_TESTS=OFF"]


TaskRegistry.get().add_task_class(Libzmq)
TaskRegistry.get().add_task_class(Cppzmq)



from jolt import attributes
from jolt.plugins import autotools, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
class M4(autotools.Autotools):
    name = "m4"
    version = "1.4.20"
    requires_src = ["fetch:alias=src,url=http://ftpmirror.gnu.org/m4/m4-{version}.tar.xz"]
    srcdir = "{fetch[src]}/m4-{version}"


TaskRegistry.get().add_task_class(M4)



from jolt import attributes, BooleanParameter, Parameter, Task
from jolt.plugins import cmake, git
from jolt.tasks import TaskRegistry
import os


@attributes.requires("requires_git")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
class Abseil(cmake.CMake):
    """ Abseil C++ Common Libraries """

    name = "abseil"
    version = Parameter("20250814.1")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/abseil/abseil-cpp.git,rev={version}"]
    srcdir = "{git[abseil-cpp]}"
    options = [
        "ABSL_MSVC_STATIC_RUNTIME=OFF",
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CMAKE_CXX_STANDARD=17",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.msvcrt = "Dynamic"
        artifact.cxxinfo.incpaths.append("include")

        with tools.cwd(artifact.path):
            for libdir in ["lib", "lib32", "lib64"]:
                if not tools.exists(libdir):
                    continue
                with tools.cwd(libdir):
                    artifact.cxxinfo.libpaths.append(libdir)
                    if self.shared:
                        artifact.environ.LD_LIBRARY_PATH.append(libdir)
                    for libfile in tools.glob("*.lib"):
                        libname, _ = os.path.splitext(libfile)
                        artifact.cxxinfo.libraries.append(libname)
                    for libfile in tools.glob("lib*.a"):
                        libname, _ = os.path.splitext(libfile)
                        artifact.cxxinfo.libraries.append(libname[3:])


TaskRegistry.get().add_task_class(Abseil)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@pkgconfig.requires()
class DBus(cmake.CMake):
    name = "dbus"
    version = Parameter("1.16.2", help="DBus version.")
    requires_git = ["git:url=https://gitlab.freedesktop.org/dbus/dbus.git,rev=dbus-{version}"]
    srcdir = "{git[dbus]}"


TaskRegistry.get().add_task_class(DBus)



from jolt import attributes, Parameter
from jolt.plugins import autotools, git, libtool, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@autotools.requires(libtool=False)
@libtool.relocate()
@libtool.requires()
@pkgconfig.requires()
@pkgconfig.to_cxxinfo("libfastjson")
class Libfastjson(autotools.Autotools):
    """ Builds and publishes liblibraries and headers. """

    name = "libfastjson"
    version = Parameter("1.2304.0", help="Libfastjson version.")
    requires_git = ["git:url=https://github.com/rsyslog/libfastjson.git,rev=v{version}"]
    srcdir = "{git[libfastjson]}"


TaskRegistry.get().add_task_class(Libfastjson)



from jolt import attributes, BooleanParameter, Parameter
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class XercesC(cmake.CMake):
    name = "xerces-c"
    version = Parameter("3.3.0", help="Xerces-C version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/apache/xerces-c.git,rev=v{version}"]
    srcdir = "{git[xerces-c]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(XercesC)



import importlib_metadata
import os

from jolt import config
from jolt import filesystem as fs
from jolt import influence
from jolt import log
from jolt import common_pb2 as common_pb
from jolt import utils
from jolt import version
from jolt.cache import ArtifactCache
from jolt.error import raise_error_if
from jolt.graph import GraphBuilder
from jolt.loader import JoltLoader
from jolt.scheduler import JoltEnvironment
from jolt.scheduler import LocalExecutor
from jolt.scheduler import LocalExecutorFactory
from jolt.tasks import Task, TaskRegistry


log.verbose("[SelfDeploy] Loaded")

_path = fs.path.dirname(__file__)
_path = fs.path.dirname(_path)
_path = fs.path.dirname(_path)


@influence.files(fs.path.join(_path, "**", "*.py"))
@influence.files(fs.path.join(_path, "**", "*.sh"))
@influence.files(fs.path.join(_path, "**", "*.xslt"))
@influence.files(fs.path.join(_path, "**", "*.template"))
@influence.attribute("dependencies")
@influence.attribute("extra_dependencies")
class Jolt(Task):
    name = "jolt"

    def __init__(self, *args, **kwargs):
        super(Jolt, self).__init__(*args, **kwargs)
        self.influence.append(
            influence.StringInfluence(
                config.get("selfdeploy", "requires", "")))
        for e in self.extra_files:
            self.influence.append(influence.DirectoryInfluence(e))

    def _verify_influence(self, *args, **kwargs):
        pass  # FIXME: Validation doesn't work properly for directories

    @property
    def loaderdir(self):
        return JoltLoader.get().joltdir

    @property
    def extra_files(self):
        ext = config.get("selfdeploy", "extra", "")
        return [fs.path.join(self.loaderdir, e) for e in ext.split(",")] if ext else []

    @property
    def extra_dependencies(self):
        return get_extra_dependencies()

    @property
    def dependencies(self):
        """
        Queries local pip database for packages and versions used by jolt.

        The data is used to pin dependencies at specific versions so that
        the virtalenv installed on a worker matches what the client had
        installed locally.

        If pip is not available or its internal implementation has changed,
        no version pinning will be performed. Instead, workers will install
        Jolt with its default loose version requirements.
        """
        return get_dependencies(["jolt"] + self.extra_dependencies)

    def publish(self, artifact, tools):
        with tools.cwd(tools.builddir()):
            pinned_reqs = self.dependencies
            if pinned_reqs:
                tools.write_file("requirements.txt", "\n".join(pinned_reqs))
                artifact.collect("requirements.txt")
        with tools.cwd(_path):
            artifact.collect('README.rst')
            artifact.collect('setup.py')
            artifact.collect('jolt/*.py')
            artifact.collect('jolt/*.sh')
            artifact.collect('jolt/*/*.py')
            artifact.collect('jolt/*/*/*.py')
            artifact.collect('jolt/*/*.xslt')
            artifact.collect('jolt/*/*.template')
            artifact.collect('jolt/bin')
            artifact.collect('jolt/plugins/selfdeploy/README.rst', flatten=True)
            artifact.collect('jolt/plugins/selfdeploy/setup.py', flatten=True)
            for e in self.extra_files:
                with tools.cwd(fs.path.dirname(e)):
                    artifact.collect(fs.path.basename(e))


@utils.cached.method
def get_dependencies(packages=None):
    reqs = packages or ["jolt"]
    pkgs = {}
    skip = set()

    while reqs:
        req = reqs.pop()

        try:
            dist = importlib_metadata.distribution(req)
        except (ImportError, importlib_metadata.PackageNotFoundError):
            dist = None
        except Exception:
            dist = None
        if dist is None:
            skip.add(req)
            continue

        for dep in dist.requires or []:
            dep = dep.split(" ", 1)[0].strip()
            dep = dep.split("[", 1)[0].strip()
            dep = dep.split(";", 1)[0].strip()
            dep = dep.split(">", 1)[0].strip()
            dep = dep.split("=", 1)[0].strip()
            dep = dep.split("<", 1)[0].strip()
            dep = dep.split("!", 1)[0].strip()
            if dep not in pkgs and dep not in skip:
                reqs.append(dep)

        pkgs[req] = f"{dist.name}=={dist.version}"

    try:
        del pkgs["jolt"]
    except KeyError:
        pass

    return list(sorted(pkgs.values()))


@utils.cached.method
def get_extra_dependencies():
    req = config.get("selfdeploy", "requires", "")
    return req.split() if req else []


@utils.cached.method
def publish_artifact():
    registry = TaskRegistry()
    registry.add_task_class(Jolt)
    acache = ArtifactCache.get()
    env = JoltEnvironment(cache=acache, queue=None)
    gb = GraphBuilder(registry, acache)
    dag = gb.build(["jolt"])
    task = dag.select(lambda graph, task: True)
    assert len(task) == 1, "Too many selfdeploy tasks found"
    task = task[0]
    if not task.is_available_remotely(cache=False):
        factory = LocalExecutorFactory()
        executor = LocalExecutor(factory, task, force_upload=True)
        executor.run(env)
    jolt_url = acache.location(task.artifacts[0])
    raise_error_if(not jolt_url, "Failed to deploy jolt to a remote cache")
    cacheUrl = config.get("http", "uri", config.get("cache", "http_uri", config.get("cache", "uri", "")) + "/files")
    substituteUrl = config.get("selfdeploy", "baseUri")
    if cacheUrl and substituteUrl:
        return task.identity, jolt_url.replace(cacheUrl, substituteUrl)
    return task.identity, jolt_url


def get_floating_version():
    identity, url = publish_artifact()
    return common_pb.Client(
        identity=identity,
        url=url,
        version=version.__version__,
    )


def get_pinned_version():
    return common_pb.Client(
        requirements=get_extra_dependencies(),
        version=version.__version__,
    )


def get_nix_version():
    return common_pb.Client(
        requirements=get_extra_dependencies(),
        version=version.__version__,
        nix=True,
    )


def get_client():
    # Floating version is a special case where we want to deploy the Jolt
    # source code to a remote cache and use that URL as the client URL
    # for workers.
    floating = config.getboolean("selfdeploy", "floating", False)
    if floating:
        return get_floating_version()

    # If Nix has been explicitly disabled, we want to pin versions.
    if not config.getboolean("selfdeploy", "nix", True):
        return get_pinned_version()

    # If Nix has been explicitly enabled, we want to use the Nix shell.
    if config.getboolean("selfdeploy", "nix", False):
        return get_nix_version()

    # If we are in a Nix shell, we want to use the Nix shell.
    if os.environ.get("IN_NIX_SHELL"):
        return get_nix_version()

    # Default to pinned version
    return get_pinned_version()



from jolt import attributes, Parameter
from jolt.pkgs import help2man, m4, texinfo
from jolt.plugins import git, autotools
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_help2man")
@attributes.requires("requires_m4")
@attributes.requires("requires_texinfo")
class Autoconf(autotools.Autotools):
    name = "autoconf"
    version = Parameter("2.72", help="Autoconf version.")
    requires_git = ["fetch:alias=src,url=https://ftpmirror.gnu.org/gnu/autoconf/autoconf-{version}.tar.gz"]
    requires_help2man = ["help2man"]
    requires_m4 = ["m4"]
    requires_texinfo = ["texinfo"]
    srcdir = "{fetch[src]}/autoconf-{version}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        self.unpack(artifact, tools)

    def unpack(self, artifact, tools):
        with tools.cwd(artifact.path, "bin"):
            for file in tools.glob("*"):
                tools.replace_in_file(
                    file,
                    artifact.strings.install_prefix,
                    artifact.final_path)

        with tools.cwd(artifact.path, "share/autoconf"):
            tools.replace_in_file(
                "autom4te.cfg",
                artifact.strings.install_prefix,
                artifact.final_path)

        artifact.strings.install_prefix = artifact.final_path


TaskRegistry.get().add_task_class(Autoconf)



from jolt import attributes, Parameter
from jolt.plugins import git, autotools, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.arch
@attributes.system
@autotools.requires()
@pkgconfig.to_cxxinfo(["libunwind", "libunwind-generic", "libunwind-coredump", "libunwind-ptrace", "libunwind-setjmp"])
class LibUnwind(autotools.Autotools):
    name = "libunwind"
    version = Parameter("1.8.3", help="libunwind version.")
    options = ["--disable-tests"]
    requires_git = ["git:url=https://github.com/libunwind/libunwind.git,rev=v{version}"]
    srcdir = "{git[libunwind]}"

    def run(self, deps, tools):
        with tools.environ(TARGET="{target}"):
            super().run(deps, tools)

    @property
    def target(self):
        platform = self.tools.expand("{system}-{arch}")
        if platform == "linux-x86_64":
            return "x86_64-linux-gnu"
        elif platform == "linux-aarch64":
            return "aarch64-linux-gnu"
        elif platform == "linux-arm":
            return "arm-linux-gnueabihf"
        else:
            self.error(f"Unsupported platform: {platform}")


TaskRegistry.get().add_task_class(LibUnwind)



from jolt import Parameter, attributes
from jolt.plugins import autotools, fetch, pkgconfig
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
@pkgconfig.to_cxxinfo(["readline"])
class Readline(autotools.Autotools):
    name = "readline"
    version = Parameter("8.2", help="Readline version.")
    requires_src = ["fetch:alias=src,url=https://ftpmirror.gnu.org/gnu/readline/readline-{version}.tar.gz"]
    srcdir = "{fetch[src]}/readline-{version}"


TaskRegistry.get().add_task_class(Readline)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake, ssl
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry
import os


@attributes.requires("requires_git")
@attributes.requires("requires_ssl")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Poco(cmake.CMake):
    name = "poco"
    version = Parameter("1.14.2", help="Poco version")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/pocoproject/poco.git,rev=poco-{version}-release"]
    requires_ssl = ["virtual/ssl"]
    srcdir = "{git[poco]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "POCO_MT=OFF",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            artifact.cxxinfo.msvcrt = "Dynamic"

        for libdir in ["lib", "lib32", "lib64"]:
            full_libdir = os.path.join(artifact.path, libdir)
            if not os.path.isdir(full_libdir):
                continue
            with tools.cwd(artifact.path, libdir):
                for libfile in tools.glob("*.lib"):
                    libname, _ = os.path.splitext(libfile)
                    artifact.cxxinfo.libraries.append(libname)
                for libfile in tools.glob("lib*.a"):
                    libname, _ = os.path.splitext(libfile)
                    artifact.cxxinfo.libraries.append(libname[3:])


TaskRegistry.get().add_task_class(Poco)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import nasm
from jolt.plugins import cmake, cxxinfo, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_nasm_{system}")
@attributes.common_metadata()
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["ssl", "crypto"])
class BoringSSL(cmake.CMake):
    name = "boringssl"
    version = Parameter("0.20251124.0", help="boringssl version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/google/boringssl.git,rev={version}"]
    requires_nasm_windows = ["nasm"]
    srcdir = "{git[boringssl]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
    ]


TaskRegistry.get().add_task_class(BoringSSL)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import nasm, pkgconfig, xz, xorg
from jolt.plugins import git, autotools
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_{system}_libx11")
@attributes.requires("requires_nasm")
@attributes.requires("requires_xz")
@attributes.system
@autotools.requires()
class FFmpeg(autotools.Autotools):
    name = "ffmpeg"
    version = Parameter("8.0.1", help="ffmpeg version.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_git = ["git:url=https://github.com/FFmpeg/FFmpeg.git,rev=n{version}"]
    requires_linux_libx11 = ["libx11"]
    requires_nasm = ["nasm"]
    requires_xz = ["xz"]
    srcdir = "{git[FFmpeg]}"
    options = [
        "--{shared[enable,disable]}-shared",
        "--{shared[disable,enable]}-static",
    ]


TaskRegistry.get().add_task_class(FFmpeg)




from jolt import Task, Parameter, ListParameter
from jolt import attributes
from jolt import log
from jolt import utils
from jolt.error import raise_task_error_if
from jolt.plugins import podman


import os
import shutil


def linux_arch_to_container_platform(arch):
    """
    Convert Linux architecture to Podman platform.

    Args:
        arch: Linux architecture.

    Returns:
        - linux/amd64
        - linux/arm
        - linux/arm/v5
        - linux/arm64
        - linux/mips
        - linux/ppc64
        - linux/riscv64
        - linux/s390x
    """
    platforms = {
        "arm": "linux/arm",
        "arm64": "linux/arm64",
        "armv5": "linux/arm/v5",
        "mips": "linux/mips64le",
        "powerpc": "linux/ppc64le",
        "riscv": "linux/riscv64",
        "s390": "linux/s390x",
        "x86": "linux/amd64",
    }
    arch = str(arch)
    try:
        return platforms[arch]
    except KeyError:
        raise ValueError(f"Unsupported architecture {arch}")


def linux_arch_to_debian_arch(arch):
    """
    Convert Linux architecture to Debian architecture.

    Args:
        arch: Linux architecture.

    Returns:
        - amd64
        - armhf
        - arm64
        - mips
        - ppc64el
        - riscv64
        - s390x
    """
    debian_arch = {
        "amd64": "amd64",
        "arm": "armhf",
        "arm64": "arm64",
        "armv5": "armel",
        "mips": "mips",
        "powerpc": "ppc64el",
        "riscv": "riscv64",
        "s390": "s390x",
        "x86": "i386",
    }
    return debian_arch[str(arch)]


class ArchParameter(Parameter):
    """
    Linux target architecture parameter.

    Supported values:

      - amd64
      - arm
      - arm64
      - armv5
      - mips
      - powerpc
      - riscv
      - s390
      - x86

    """

    def __init__(self, *args, **kwargs):
        kwargs["values"] = [
            "amd64",
            "arm",
            "arm64",
            "armv5",
            "mips",
            "powerpc",
            "riscv",
            "s390",
            "x86",
        ]
        kwargs["help"] = "Target architecture."
        super().__init__(*args, **kwargs)


class _ContainerImageBase(podman.ContainerImage):
    """ Builds a container image from external source tree """

    abstract = True
    """ Must be subclassed """

    arch = ArchParameter()
    """ Target architecture [amd64, arm, arm64, armv5, mips, powerpc, riscv, s390, x86] """

    @property
    def target(self):
        return linux_arch_to_container_platform(self.arch)


class _DebianSdkImage(podman.ContainerImage):
    """ Provides a Debian SDK for building Linux kernel and U-Boot """

    abstract = True
    """ Must be subclassed """

    arch = ArchParameter()
    """ Target architecture """

    dockerfile = """
    FROM debian:{version}-slim
    ARG DEBIAN_FRONTEND=noninteractive
    RUN apt-get update && apt-get install -y crossbuild-essential-{debian_arch}
    """

    version = "stable"
    """ Debian codename/version """

    @property
    def debian_arch(self):
        return {
            "amd64": "amd64",
            "arm": "armhf",
            "arm64": "arm64",
            "mips": "mips",
            "powerpc": "ppc64el",
            "riscv": "riscv64",
            "s390": "s390x",
            "x86": "i386",
        }[str(self.arch)]


class DebianHostSdk(Task):
    """
    Helper task that exports a Debian host's cross-compiler toolchains to consumers.

    The task verifies that the cross-compiler toolchains are installed on the host.
    If not, the task will raise an error with instructions.

    Exported environment variables:

      - ``CC``: Cross-compiler
      - ``CPP``: Cross-preprocessor
      - ``CXX``: Cross-C++ compiler
      - ``CROSS_COMPILE``: Cross-compile prefix

    """
    abstract = True
    """ Must be subclassed """

    arch = ArchParameter()
    """ Target architecture [amd64, arm, arm64, armv5, mips, powerpc, riscv, s390, x86] """

    def publish(self, artifact, tools):
        arch_to_cross_compile = {
            "amd64": "x86_64-linux-gnu-",
            "arm": "arm-linux-gnueabihf-",
            "arm64": "aarch64-linux-gnu-",
            "armv5": "arm-linux-gnueabi-",
            "mips": "mips-linux-gnu-",
            "powerpc": "powerpc64-linux-gnu-",
            "riscv": "riscv64-linux-gnu-",
            "s390": "s390x-linux-gnu-",
            "x86": "i686-linux-gnu-",
        }
        debian_arch = linux_arch_to_debian_arch(self.arch)
        debian_cross = arch_to_cross_compile[str(self.arch)]
        raise_task_error_if(
            not tools.which(arch_to_cross_compile[str(self.arch)] + "gcc"),
            self, f"Cross compiler not found. Please install crossbuild-essential-{debian_arch} package.")
        artifact.environ.CC = f"{debian_cross}gcc"
        artifact.environ.CPP = f"{debian_cross}cpp"
        artifact.environ.CXX = f"{debian_cross}g++"
        artifact.environ.CROSS_COMPILE = debian_cross


class Initramfs(_ContainerImageBase):
    """
    Builds an initramfs image using Podman.

    The task builds the initramfs image using the given Dockerfile and
    publishes the resulting cpio archive. The task requires the following
    attributes to be set:

    - arch: Target architecture
    - dockerfile: Path to Dockerfile, or Dockerfile content.

    When building images for an architecture other than the host, the
    binfmt-support package must be installed and configured to support the
    target architecture. The package is available in most Linux distributions.

    The location of the resulting cpio archive is stored in the
    'artifact.paths.cpio' attribute.

    """
    abstract = True
    """ Must be subclassed """

    output = ["cpio"]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.strings.arch = str(self.arch)
        artifact.paths.cpio = "cpio/{_imagefile}.cpio"


class Squashfs(_ContainerImageBase):
    """
    Builds a squashfs image using Podman.

    The task builds a container image using the given Dockerfile and converts
    the resulting container filesystem to a squashfs image which is published.

    When building images for an architecture other than the host, the
    binfmt-support package must be installed and configured to support running
    applications for the target architecture. The package is available in most
    Linux distributions.

    The location of the resulting squashfs image is stored in the
    ``artifact.paths.squashfs`` artifact attribute.
    """
    abstract = True
    """ Must be subclassed """

    output = ["squashfs"]

    size = None
    """
    Size of the squashfs image.

    Typically used to align the image size to a supported SD card size (power of two).

    Supported units are 'K', 'M', 'G', 'T'.
    """

    def run(self, deps, tools):
        super().run(deps, tools)
        if self.size:
            with tools.cwd(tools.builddir("squashfs")):
                tools.run("fallocate -l {size} image.squashfs")

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.strings.arch = str(self.arch)
        artifact.paths.squashfs = "squashfs/{_imagefile}.squashfs"


class Ext4(_ContainerImageBase):
    """
    Builds an ext4 image using Podman.

    The task builds a container image using the given Dockerfile and converts
    the resulting container filesystem to an ext4 image which is published.

    When building images for an architecture other than the host, the
    binfmt-support package must be installed and configured to support running
    applications for the target architecture. The package is available in most
    Linux distributions.

    The location of the resulting squashfs image is stored in the
    ``artifact.paths.squashfs`` artifact attribute.
    """
    abstract = True
    """ Must be subclassed """

    output = ["ext4"]

    size = None
    """
    Size of the ext4 image.

    Typically used to align the image size to a supported SD card size (power of two).

    Supported units are 'K', 'M', 'G', 'T'.
    """

    def run(self, deps, tools):
        super().run(deps, tools)
        if self.size:
            with tools.cwd(tools.builddir("ext4")):
                tools.run("fallocate -l {size} image.ext4")

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.strings.arch = str(self.arch)
        artifact.paths.ext4 = "ext4/{_imagefile}.ext4"


class _KernelBase(Task):
    abstract = True
    """ Must be subclassed """

    arch = ArchParameter()
    """ Linux target architecture [amd64, arm, arm64, mips, powerpc, riscv, s390, x86] """

    defconfig = Parameter("allnoconfig", help="Name of kernel defconfig")
    """ Default configuration """

    features = ListParameter([], help="List of feature configs to apply")
    """
    List of feature configuration snippets to apply.

    If used, the task will search for feature configuration snippets in the
    configured paths and merge them into the selected defconfig.
    """

    configpaths = [
        "{srcdir}/arch/{arch}/configs",
    ]
    """
    Paths to configuration file snippets.

    When building the kernel with features, the task will search for
    feature configuration snippets in these directories. The snippets
    are applied and merged with the selected defconfig.
    """

    defconfigpath = "arch/{arch}/configs/{defconfig}_defconfig"
    """ Path to defconfig file. """

    srcdir = None
    """ Location of Linux kernel source tree """

    targets = ListParameter(help="Targets to build and publish")
    """ Override and set accepted values """

    def clean(self, tools):
        self.moddir = tools.builddir("modules")
        self.objdir = tools.builddir("objects", incremental=True)
        tools.rmtree(self.moddir)
        tools.rmtree(self.objdir)

    def run(self, deps, tools):
        raise_task_error_if(not tools.which("bison"), self, "bison is required to build the kernel")
        raise_task_error_if(not tools.which("flex"), self, "flex is required to build the kernel")
        raise_task_error_if(not tools.which("make"), self, "make is required to build the kernel")

        self.objdir = tools.builddir("objects", incremental=True)

        config = tools.expand_path("{objdir}/.config")
        tools.unlink(config, ignore_errors=True)

        arch = f"ARCH={self.arch}" if self.arch else ""
        verbose = "V=1" if log.is_verbose() else ""
        defconf_conf = self.run_find_defconfig_configs(deps, tools)
        feature_confs = self.run_find_feature_configs(deps, tools)
        defconfigpath = tools.expand(self.defconfigpath)
        if self.defconfig in ["defconfig"]:
            defconfigpath = os.path.join(os.path.dirname(defconfigpath), "defconfig")

        with tools.cwd(self.srcdir):
            if self.defconfig in ["allnoconfig"]:
                self.info("Generating allnoconfig configuration")
                tools.run("{} make allnoconfig O={objdir}", arch)
                tools.run("{} scripts/kconfig/merge_config.sh -O {objdir} {objdir}/.config {} {}", arch, defconf_conf, " ".join(feature_confs))
            else:
                self.info("Generating {defconfig} configuration")
                tools.run("{} scripts/kconfig/merge_config.sh -O {objdir} {} {} {}", arch, defconfigpath, defconf_conf, " ".join(feature_confs))

            # Run build for each requested target
            for target in self.targets:
                fn = getattr(self, f"run_{target}", None)
                assert fn, f"Dont know how to build {target}"
                with tools.runprefix("{} make {} O={objdir} -j{}", arch, verbose, tools.thread_count()):
                    fn(deps, tools)

    def run_find_defconfig_configs(self, deps, tools):
        """ Finds local overrides for the chosen defconfig """
        configpaths = [tools.expand_path(tools.expand(path)) for path in getattr(self, "configpaths", [])]
        return shutil.which(tools.expand("{defconfig}.config"), os.F_OK, os.pathsep.join(configpaths)) or ""

    def run_find_feature_configs(self, deps, tools):
        """ Finds local overrides for chosen features """
        configpaths = [tools.expand_path(tools.expand(path)) for path in getattr(self, "configpaths", [])]
        return [shutil.which(feature + ".config", os.F_OK, os.pathsep.join(configpaths)) or "" for feature in self.features]

    def publish(self, artifact, tools):
        artifact.strings.arch = str(self.arch)

        for target in self.targets:
            fn = getattr(self, f"publish_{target}", None)
            assert fn, f"Dont know how to publish {target}"
            fn(artifact, tools)


class UBoot(_KernelBase):
    """
    Builds u-boot makefile target(s) and publishes the result.

    An implementor must subclass this task and set the following attributes:

      - srcdir: Path to U-boot source tree

    It is recommended to use a task dependency to setup the
    required tools by adding them to ``PATH`` and/or setting the ``CROSS_COMPILE``
    environment variable. When building on a Debian host, the
    :class:`jolt.plugins.linux.DebianHostSdk` helper task can be used to export
    the cross-compiler tools.

    """
    abstract = True

    defconfig = Parameter("allnoconfig", help="Name of u-boot defconfig")
    """ Default configuration """

    targets = ListParameter(
        ["uboot"],
        values=["tools", "uboot"],
        help="Targets to build and publish",
    )
    """ Build targets [tools, uboot] """

    def run_tools(self, deps, tools):
        self.info("Building tools ...")
        tools.run("tools")

    def run_uboot(self, deps, tools):
        self.info("Building u-boot ...")
        tools.run("u-boot.bin")

    def publish_tools(self, artifact, tools):
        self.info("Publishing tools ...")
        with tools.cwd(self.objdir):
            artifact.collect("tools/mkimage")
            artifact.environ.PATH.append("tools")

    def publish_uboot(self, artifact, tools):
        self.info("Publishing u-boot ...")
        with tools.cwd(self.objdir):
            artifact.collect("*.bin")


@utils.concat_attributes("dtbs", "dtbs_{defconfig}")
class Kernel(_KernelBase):
    """
    Builds kernel makefile target(s) and publishes the result.

    Targets are selected by assigning the ``targets`` parameter. The following
    values are supported:

        - ``binary``: Build kernel binary
        - ``dtbs``: Build device trees
        - ``dtc``: Build device tree compiler
        - ``gzimage``: Build gzipped kernel image
        - ``image``: Build kernel image
        - ``modules``: Build kernel modules
        - ``uimage``: Build uImage
        - ``vmlinux``: Build vmlinux
        - ``zimage``: Build zImage

    Mutliple targets can be selected at once, such as ``vmlinux+dtbs+modules``.
    The resulting artifacts are published into different directories
    based on the target name.

    It is recommended to use a task dependency to setup the
    required tools by adding them to ``PATH`` and/or setting the ``CROSS_COMPILE``
    environment variable. When building on a Debian host, the
    :class:`jolt.plugins.linux.DebianHostSdk` helper task can be used to export
    the cross-compiler tools.

    The following tools are required to be available in PATH:

        - bison
        - flex
        - make
        - gcc

    Artifact path attributes are created to point to the published files.

    """

    abstract = True
    """ Must be subclassed """

    dtbs = []
    """
    List of device trees to build when the 'dtbs' target is requested.

    If empty, all device trees associated with the target architecture are built.
    """

    loadaddr = 0
    """
    Kernel load address.
    """

    loadaddr_fdt = 0
    """
    Device-tree load address.
    """

    targets = ListParameter(
        ["vmlinux"],
        values=["binary", "dtbs", "dtc", "gzimage", "image", "modules", "uimage", "vmlinux", "zimage"],
        help="Targets to build and publish",
    )
    """ Build targets [binary, dtbs, dtc, gzimage, image, modules, uimage, vmlinux, zimage] """

    def run_binary(self, deps, tools):
        self.info("Building binary kernel ...")
        tools.run("LOADADDR={loadaddr} vmlinux")
        objcopy = tools.getenv("OBJCOPY", tools.getenv("CROSS_COMPILE", "") + "objcopy")
        tools._run_prefix = []  # Hack to cancel previous prefix
        tools.run("{} -O binary -R .note -R .comment -S {objdir}/vmlinux {objdir}/vmlinux.bin", objcopy)

    def run_dtbs(self, deps, tools):
        self.info("Building device trees ...")
        if not self._dtbs():
            tools.run("dtbs")
        else:
            tools.run(" ".join(self._dtbs()))

    def run_dtc(self, deps, tools):
        self.info("Building device tree compiler ...")
        tools.run("CONFIG_DTC=y scripts")

    def run_modules(self, deps, tools):
        self.info("Building modules ...")
        self.moddir = tools.builddir("modules")
        tools.run("INSTALL_MOD_PATH={moddir} modules")
        tools.run("INSTALL_MOD_PATH={moddir} modules_install")

    def run_image(self, deps, tools):
        self.info("Building Image ...")
        tools.run("LOADADDR={loadaddr} Image")

    def run_gzimage(self, deps, tools):
        self.info("Building Image.gz ...")
        tools.run("LOADADDR={loadaddr} Image.gz")

    def run_uimage(self, deps, tools):
        self.info("Building uImage ...")
        tools.run("LOADADDR={loadaddr} uImage")

    def run_vmlinux(self, deps, tools):
        self.info("Building vmlinux ...")
        tools.run("vmlinux")

    def run_zimage(self, deps, tools):
        self.info("Building zImage ...")
        tools.run("zImage")

    def publish(self, artifact, tools):
        super().publish(artifact, tools)

        if self.loadaddr != 0:
            artifact.strings.loadaddr = str(self.loadaddr)
        if self.loadaddr_fdt != 0:
            artifact.strings.loadaddr_fdt = str(self.loadaddr_fdt)

    def publish_binary(self, artifact, tools):
        self.info("Publishing binary ...")
        with tools.cwd(self.objdir):
            artifact.collect("vmlinux.bin", "binary/")
            artifact.paths.binary = "binary/vmlinux.bin"

    def publish_dtbs(self, artifact, tools):
        self.info("Publishing device trees ...")
        with tools.cwd(self.objdir, "arch/{arch}/boot/dts"):
            if not self._dtbs():
                artifact.collect("**/*.dtb", "dtbs/")
            else:
                for dtb in self._dtbs():
                    artifact.collect(dtb, "dtbs/")
            artifact.paths.dtbs = "dtbs"

    def publish_dtc(self, artifact, tools):
        self.info("Publishing device tree compiler ...")
        with tools.cwd(self.objdir, "scripts/dtc"):
            artifact.collect("dtc", "bin/")
            artifact.environ.PATH.append("bin")

    def publish_modules(self, artifact, tools):
        self.info("Publishing modules ...")
        with tools.cwd(self.moddir):
            artifact.collect(".", "modules/", symlinks=True)
            artifact.paths.modules = "modules"

        with tools.cwd(self.objdir):
            if os.path.exists(tools.expand_path("Module.symvers")):
                artifact.collect("Module.symvers")
                artifact.paths.symvers = "Module.symvers"

    def publish_image(self, artifact, tools):
        self.info("Publishing Image ...")
        with tools.cwd(self.objdir, "arch/{arch}/boot"):
            artifact.collect("Image", "image/")
            artifact.paths.image = "image/Image"

    def publish_gzimage(self, artifact, tools):
        self.info("Publishing Image.gz ...")
        with tools.cwd(self.objdir, "arch/{arch}/boot"):
            artifact.collect("Image.gz", "gzimage/")
            artifact.paths.gzimage = "gzimage/Image.gz"

    def publish_uimage(self, artifact, tools):
        self.info("Publishing uImage ...")
        with tools.cwd(self.objdir, "arch/{arch}/boot"):
            artifact.collect("uImage", "uimage/")
            artifact.paths.uimage = "uimage/uImage"

    def publish_vmlinux(self, artifact, tools):
        self.info("Publishing vmlinux ...")
        with tools.cwd(self.objdir):
            artifact.collect("vmlinux", "vmlinux/")
            artifact.paths.vmlinux = "vmlinux/vmlinux"

    def publish_zimage(self, artifact, tools):
        self.info("Publishing zImage ...")
        with tools.cwd(self.objdir, "arch/{arch}/boot"):
            artifact.collect("zImage", "zimage/")
            artifact.paths.zimage = "zimage/zImage"


class Module(Kernel):
    """
    Builds a kernel module from external source tree.

    The task is based on the :class:`Kernel` task and builds a kernel module
    from the given source tree specified by the ``srcdir_module`` attribute.
    The ``targets`` attribute is set to ``modules`` by default.

    See :class:`Kernel` for additional information on building kernel targets.
    In particular, the ``srcdir`` attribute must be set to the kernel source tree
    to build the modules against.
    """

    abstract = True
    """ Must be subclassed """

    srcdir_module = None
    """ Path to kernel module source tree """

    targets = ["modules"]
    """ Build targets [modules] """

    def run(self, deps, tools):
        self.targets = ["modules"]
        super().run(deps, tools)

    def run_modules(self, deps, tools):
        with tools.cwd(self.objdir):
            for _, artifact in deps.items():
                if str(artifact.paths.symvers):
                    self.info("Copying Module.symvers")
                    tools.copy(str(artifact.paths.symvers), ".")

        self.info("Building modules ...")
        self.moddir = tools.builddir("modules")
        tools.run("INSTALL_MOD_PATH={moddir} modules_prepare")
        tools.run("INSTALL_MOD_PATH={moddir} M={joltdir}/{srcdir_module} modules")
        tools.run("INSTALL_MOD_PATH={moddir} M={joltdir}/{srcdir_module} modules_install")


@utils.concat_attributes("configs", "configs")
class FIT(Task):
    """
    Builds and publishes a FIT Image.

    The task builds a FIT image using the given kernel, device tree and ramdisk
    tasks and publishes the resulting image. Multiple FIT configurations can be
    created by defining the ``configs`` attribute. The image is optinally signed
    using the provided key.

    The task requires the ``mkimage`` tool to be available in PATH.
    """

    abstract = True
    """ Must be subclassed """

    configs = {}
    """
    FIT configurations to create.

    Dictionary of 2 or 3 element tuples according to this format:

    .. code-block:: python

      configs = {
        "<name of config>": ("<name of kernel task>", "<name of dtb>"),
        "<name of config with ramdisk>": ("<name of kernel task>", "<name of dtb>", "<name of ramdisk task>"),
      }

    Example:

    .. code-block:: python

      requires = [
        "kernel=linux/kernel:arch=arm,targets=dtbs+binary_gz,defconfig=bcm2835",
        "ramdisk=busybox/irfs:arch=armv7"
      ]

      configs = {
        "conf-rpi0w": ("kernel", "bcm2835-rpi-zero-w.dtb"),
        "conf-rpi0w-irfs": ("kernel", "bcm2835-rpi-zero-w.dtb", "ramdisk"),
      }

    Loaded from defs/{defconfig}.py
    """

    config_default = None
    """ Default FIT configuration """

    _template = """/dts-v1/;

/ {
    description = "U-Boot fitImage";

    images {
        {% for i, kernel in enumerate(_kernels) %}
        kernel@{{kernel}} {
            description = "Linux kernel";
            {% if str(deps[kernel].paths.zimage) %}
            data = /incbin/("{{deps[kernel].paths.zimage}}");
            compression = "none";
            {% elif str(deps[kernel].paths.gzimage) %}
            data = /incbin/("{{deps[kernel].paths.gzimage}}");
            compression = "gzip";
            {% elif str(deps[kernel].paths.image) %}
            data = /incbin/("{{deps[kernel].paths.image}}");
            compression = "none";
            {% else %}
            #error "Kernel type not supported."
            {% endif %}
            type = "kernel";
            arch = "{{deps[kernel].strings.arch}}";
            os = "linux";
            {% if deps[kernel].strings.loadaddr %}
            load = <{{deps[kernel].strings.loadaddr}}>;
            entry = <{{deps[kernel].strings.loadaddr}}>;
            {% endif %}
            hash {
                algo = "sha256";
            };
        };

        {% endfor %}
        {% for kernel, dtb in _dtbs %}
        fdt@{{path.basename(dtb)}} {
            description = "Flattened Device Tree blob";
            data = /incbin/("{{deps[kernel].paths.dtbs}}/{{dtb}}");
            type = "flat_dt";
            arch = "{{deps[kernel].strings.arch}}";
            compression = "none";
            os = "linux";
            {% if deps[kernel].strings.loadaddr_fdt %}
            load = <{{deps[kernel].strings.loadaddr_fdt}}>;
            {% endif %}
            hash {
                    algo = "sha256";
            };
        };

        {% endfor %}
        {% for ramdisk in _ramdisks %}
        ramdisk@{{ramdisk}} {
            description = "Ramdisk";
            data = /incbin/("{{deps[ramdisk].paths.cpio}}");
            type = "ramdisk";
            arch = "{{deps[ramdisk].strings.arch}}";
            os = "linux";
            {% if deps[ramdisk].strings.compression %}
            compression = "{{deps[ramdisk].strings.compression}}";
            {% else %}
            compression = "none";
            {% endif %}
            hash {
                algo = "sha256";
            };
        };

        {% endfor %}
    };

    configurations {
        {% if config_default %}
        default = "conf@{{config_default}}";
        {% endif %}

        {% for name, config in _configs().items() %}
        conf@{{name}} {
                {% if len(config) > 2 %}
                description = "Linux kernel, FDT blob, Ramdisk";
                {% else %}
                description = "Linux kernel, FDT blob";
                {% endif %}
                kernel = "kernel@{{config[0]}}";
                os = "linux";
                fdt = "fdt@{{path.basename(config[1])}}";
                {% if len(config) > 2 %}
                ramdisk = "ramdisk@{{config[2]}}";
                {% endif %}

                hash {
                        algo = "sha256";
                };
                signature {
                        algo = "sha256,rsa2048";
                        {% if signature_key_name %}
                        key-name-hint = "{{signature_key_name}}";
                        {% endif %}
                        {% if len(config) > 2 %}
                        sign-images = "kernel", "fdt", "ramdisk";
                        {% else %}
                        sign-images = "kernel", "fdt";
                        {% endif %}
                };
        };

        {% endfor %}
    };
};
    """

    signature_key_name = None
    """ Name of key used to sign image. If None, image won't be signed. """

    signature_key_path = None
    """ Directory path to signature key """

    def run(self, deps, tools):
        assert tools.which("mkimage"), "mkimage is required to build the FIT image"

        kernels = []
        dtbs = []
        ramdisks = []

        # Add kernels and dtbs to individual lists
        for name, config in self._configs().items():
            kernels.append(config[0])
            dtbs.append((config[0], config[1]))
            try:
                ramdisks.append(config[2])
            except IndexError:
                pass
        dtbs = list(set(dtbs))
        kernels = list(set(kernels))

        # Render image device tree source template
        self.info("Rendering FIT image source file")
        its = tools.render(
            self._template,
            enumerate=enumerate,
            len=len,
            path=os.path,
            print=print,
            str=str,
            deps=deps,
            _dtbs=list(set(dtbs)),
            _kernels=list(set(kernels)),
            _ramdisks=list(set(ramdisks)))

        with tools.cwd(tools.builddir()):
            tools.write_file("fitImage.its", its, expand=False)
            print(its)

            self.info("Building FIT image")
            tools.run(["mkimage", "-V"], shell=False)
            tools.run(["mkimage", "-D", "-I dts -O dtb -p 2000", "-f", "fitImage.its", "fitImage"], shell=False)

            # Optionally sign the image
            if self.signature_key_name:
                self.info("Signing FIT image")
                signature_key_path = tools.expand_path(self.signature_key_path)
                tools.copy("fitImage", "fitImage.unsigned")
                tools.run(["mkimage", "-D", "-I dts -O dtb -p 2000", "-F", "-k", signature_key_path, "-r", "fitImage"], shell=False)

    def publish(self, artifact, tools):
        with tools.cwd(tools.builddir()):
            artifact.collect("fitImage")
            artifact.collect("fitImage.unsigned")
            artifact.collect("fitImage.its")
            artifact.paths.fitimage = "fitImage"


@attributes.attribute("binary", "binary_{arch}")
class Qemu(Task):
    """
    Runs Qemu with the given kernel, initramfs and rootfs.

    The task automatically selects the correct Qemu binary for the architecture
    and builds the command line arguments based on the provided kernel, initramfs
    and rootfs tasks.

    Additional arguments can be passed to QEMU by setting the ``arguments``
    attribute.

    The selected Qemu binary must be available in PATH.
    """

    abstract = True
    """ Must be subclassed """

    arch = ArchParameter()
    """ Target architecture [amd64, arm, arm64, mips, powerpc, riscv, s390, x86] """

    arguments = []
    """ Additional arguments to pass to QEMU """

    binary = None
    """ Name of QEMU binary """

    binary_arm = "qemu-system-arm"
    binary_arm64 = "qemu-system-aarch64"
    binary_mips = "qemu-system-mips"
    binary_powerpc = "qemu-system-ppc"
    binary_riscv = "qemu-system-riscv64"
    binary_s390 = "qemu-system-s390x"
    binary_x86 = "qemu-system-x86_64"

    cacheable = False
    """ Task is not cacheable """

    dtb = None
    """ Path to device tree in kernel artifact """

    initrd = None
    """ Name of initrd/ramfs task, if any """

    kernel = None
    """ Name of kernel task """

    machine = None
    """ Target machine """

    memory = None
    """ Memory size. Example: 512M """

    rootfs = None
    """ Name of rootfs task, if any """

    def requires(self):
        r = []
        if self.kernel:
            r.append("kernel=" + self.kernel)
        if self.initrd:
            r.append("initrd=" + self.initrd)
        if self.rootfs:
            r.append("rootfs=" + self.rootfs)
        return r

    def run(self, deps, tools):
        assert tools.which(self.binary), "{binary} is required to run QEMU"
        assert self.machine, "Machine type is required to run QEMU"

        self.deps = deps

        # Get kernel and initrd paths
        arguments = self.expand(self.arguments)
        dtb = ["-dtb", os.path.join(str(deps[self.kernel].paths.dtbs), self.dtb)] if self.dtb else []
        initrd = ["-initrd", str(deps[self.initrd].paths.cpio)] if self.initrd else []
        kernel = ["-kernel", str(deps[self.kernel].paths.zimage)] if self.kernel else []
        memory = ["-m", self.memory] if self.memory else []

        # Run QEMU
        binary = tools.which(self.binary)

        cmdline = [
            binary,
            "-M", self.machine,
        ] + kernel + dtb + initrd + memory + arguments

        self.info("Running QEMU with command: {}", cmdline)

        os.execv(binary, cmdline)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import ninja
from jolt.plugins import git
from jolt.plugins.ninja import CXXLibrary
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_ninja")
class Lua(CXXLibrary):
    name = "lua"
    version = Parameter("5.4.8", help="Lua version.")
    shared = BooleanParameter(False, help="Build shared libraries.")

    requires_git = ["git:url=https://github.com/lua/lua.git,rev=v{version},hash=true"]
    requires_ninja = ["ninja"]
    sources = [
        "{git[lua]}/lapi.c",
        "{git[lua]}/lauxlib.c",
        "{git[lua]}/lbaselib.c",
        "{git[lua]}/lcode.c",
        "{git[lua]}/lcorolib.c",
        "{git[lua]}/lctype.c",
        "{git[lua]}/ldblib.c",
        "{git[lua]}/ldebug.c",
        "{git[lua]}/ldo.c",
        "{git[lua]}/ldump.c",
        "{git[lua]}/lfunc.c",
        "{git[lua]}/lgc.c",
        "{git[lua]}/linit.c",
        "{git[lua]}/liolib.c",
        "{git[lua]}/llex.c",
        "{git[lua]}/lmathlib.c",
        "{git[lua]}/lmem.c",
        "{git[lua]}/loadlib.c",
        "{git[lua]}/lobject.c",
        "{git[lua]}/lopcodes.c",
        "{git[lua]}/loslib.c",
        "{git[lua]}/lparser.c",
        "{git[lua]}/lstate.c",
        "{git[lua]}/lstring.c",
        "{git[lua]}/lstrlib.c",
        "{git[lua]}/ltable.c",
        "{git[lua]}/ltablib.c",
        "{git[lua]}/ltests.c",
        "{git[lua]}/ltm.c",
        "{git[lua]}/lundump.c",
        "{git[lua]}/lutf8lib.c",
        "{git[lua]}/lvm.c",
        "{git[lua]}/lzio.c",
    ]


TaskRegistry.get().add_task_class(Lua)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class Glslang(cmake.CMake):
    name = "glslang"
    version = Parameter("16.1.0", help="Glslang version.")
    requires_git = ["git:url=https://github.com/KhronosGroup/glslang.git,rev={version}"]

    def run(self, deps, tools):
        with tools.cwd("{git[glslang]}"):
            tools.run("python3 update_glslang_sources.py")
            super().run(deps, tools)


TaskRegistry.get().add_task_class(Glslang)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import brotli, ssl, zlib, zstd
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_brotli")
@attributes.requires("requires_git")
@attributes.requires("requires_ssl")
@attributes.requires("requires_zlib")
@attributes.requires("requires_zstd")
@attributes.system
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Curl(cmake.CMake):
    name = "curl"
    version = Parameter("8.17.0", help="Curl version.")
    pic = BooleanParameter(False, help="Build with position independent code.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_brotli = ["brotli:pic={pic},shared={shared}"]
    requires_git = ["git:url=https://github.com/curl/curl.git,rev=curl-{_version_tag},submodules=true"]
    requires_ssl = ["virtual/ssl:pic={pic},shared={shared}"]
    requires_zlib = ["virtual/zlib:pic={pic},shared={shared}"]
    requires_zstd = ["zstd:pic={pic},shared={shared}"]
    srcdir = "{git[curl]}"
    options = [
        "BUILD_SHARED_LIBS={shared[ON,OFF]}",
        "CURL_USE_LIBPSL=OFF",
        "USE_LIBIDN2=OFF",
        "OPENSSL_USE_STATIC_LIBS={shared[OFF,ON]}",
        "CMAKE_POSITION_INDEPENDENT_CODE={pic[ON,OFF]}",
    ]

    @property
    def _version_tag(self):
        return str(self.version).replace('.', '_')

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.system == "windows":
            if self.shared:
                artifact.cxxinfo.libraries.append("libcurl_imp")
            else:
                artifact.cxxinfo.libraries.append("libcurl")
        else:
            artifact.cxxinfo.libraries.append("curl")


TaskRegistry.get().add_task_class(Curl)



from jolt import attributes, BooleanParameter, Parameter
from jolt.tasks import TaskRegistry
from jolt.plugins import cxxinfo, fetch, ninja


@attributes.requires("requires_src")
@attributes.system
@ninja.attributes.cflags("cflags_linux_pic_{pic[on,off]}")
class LibBzip2(ninja.CXXLibrary):
    name = "libbzip2"
    binary = "bzip2"
    version = Parameter("1.0.8", help="bzip2 version.")
    pic = BooleanParameter(True, help="Build position independent code")
    shared = BooleanParameter(False, help="Build shared libraries")
    cstd = 90
    cflags_darwin_pic_on = ["-fPIC"]
    cflags_linux_pic_on = ["-fPIC"]
    incremental = False
    requires_src = ["fetch:alias=src,url=https://sourceware.org/pub/bzip2/bzip2-{version}.tar.gz"]
    source_influence = False
    sources = [
        "{fetch[src]}/bzip2-{version}/blocksort.c",
        "{fetch[src]}/bzip2-{version}/bzlib.c",
        "{fetch[src]}/bzip2-{version}/bzlib.h",
        "{fetch[src]}/bzip2-{version}/compress.c",
        "{fetch[src]}/bzip2-{version}/crctable.c",
        "{fetch[src]}/bzip2-{version}/decompress.c",
        "{fetch[src]}/bzip2-{version}/huffman.c",
        "{fetch[src]}/bzip2-{version}/randtable.c",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.cxxinfo.incpaths.append("include")
        artifact.environ.CMAKE_PREFIX_PATH.append(".")

        with tools.cwd("{fetch[src]}/bzip2-{version}"):
            artifact.collect("bzlib.h", "include/")


TaskRegistry.get().add_task_class(LibBzip2)



from concurrent.futures import ThreadPoolExecutor, as_completed, Future
import copy
from functools import wraps
import os
import queue
from threading import Lock

from jolt import common_pb2 as common_pb
from jolt import config
from jolt import hooks
from jolt import log
from jolt import utils
from jolt import tools
from jolt.error import raise_task_error
from jolt.error import raise_task_error_if
from jolt.graph import PruneStrategy
from jolt.options import JoltOptions
from jolt.timer import Timer


class JoltEnvironment(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TaskQueue(object):
    """
    A helper class for tracking tasks in progress and their completion.
    """

    def __init__(self):
        self.futures = {}
        self.futures_lock = Lock()
        self.duration_acc = utils.duration_diff(0)
        self._aborted = False
        self._timer = Timer(60, self._log_task_running_time)
        self._timer.start()

    def _log_task_running_time(self):
        with self.futures_lock:
            for future in self.futures:
                self.futures[future].task.log_running_time()

    def submit(self, executor):
        """
        Submit an exeuctor to the task queue for execution.

        The method schedules the executor for execution and returns a Future object
        that may be used to track completion of the task.
        """

        if self._aborted:
            return None

        env = JoltEnvironment(queue=self)
        future = executor.schedule(env)
        with self.futures_lock:
            self.futures[future] = executor
        return future

    def wait(self):
        """
        Wait for any task to complete.

        The method waits for the next task to complete and returns the task and any
        exception that may have occurred during execution. If no task is in progress,
        the method returns None, None.
        """

        for future in as_completed(self.futures):
            task = self.futures[future].task
            try:
                future.result()
            except Exception as error:
                return task, error
            finally:
                self.duration_acc += task.duration_running or 0
                with self.futures_lock:
                    del self.futures[future]
            return task, None
        return None, None

    def abort(self):
        """
        Abort all tasks in progress.

        The method cancels all tasks in progress and prevents any new tasks from being
        submitted to the task queue. The method doesn't wait for all tasks to complete
        before returning.
        """
        self._aborted = True
        with self.futures_lock:
            for future, executor in self.futures.items():
                executor.cancel()
                future.cancel()
            if len(self.futures):
                log.info("Waiting for tasks to finish, please be patient")
        self._timer.cancel()

    def shutdown(self):
        """
        Shutdown the task queue.
        """
        self._timer.cancel()

    def is_aborted(self):
        """ Returns true if the task queue has been aborted. """
        return self._aborted

    def in_progress(self, task):
        """ Returns true if the task is in progress. """
        with self.futures_lock:
            return task in self.futures.values()

    def empty(self):
        """ Returns true if the task queue is empty. """
        with self.futures_lock:
            return len(self.futures) == 0


class Executor(object):
    """
    Base class for all executors.

    An executor is responsible for running a task. It is created by an executor
    factory and is submitted to a task queue. The factory is also
    responsible for hosting a thread pool that will run the executors it creates.

    The type of executor created by the factory depends on the execution strategy
    selected by the user through command line options. The strategy is responsible
    for deciding which executor to create for each task.

    An implementation of an executor must implement the run method, which is called
    from the thread pool. The run method is responsible for running the task and
    handling any exceptions that may occur during execution.
    """

    def __init__(self, factory):
        self.factory = factory

    def schedule(self, env):
        """ Schedule the task for execution.

        This method is called by the task queue to schedule the task for execution
        in the factory thread pool. The method must return a Future object that
        represents the task execution. The Future object is used to track the
        execution of the task and to retrieve the result of the execution
        once it is completed.

        The method must be implemented by all executors. They must call the
        factory submit method to schedule the task for execution and also
        mark the task as in progress with set_in_progress().

        Args:
            env: The JoltEnvironment object that contains the queue and cache objects.

        """
        return self.factory.submit(self, env)

    def cancel(self):
        """
        Cancel the task.

        This method is optional and may be implemented by executors that support
        cancellation of tasks, such as network executors where a remote scheduler
        may be able to cancel a task that is already running.

        By default, the method does nothing.
        """
        pass

    def is_aborted(self):
        """ Check if executor has been aborted. """
        return self.factory.is_aborted()

    def run(self, env):
        """
        Run the task.

        This method must be implemented by all executors. It is called from the
        factory thread pool and is responsible for running the task
        and handling any exceptions that may occur during execution.
        Any exceptions raised by the task must, if caught, be re-raised to
        the caller unless the task is marked as unstable, in which case the
        exception should be logged and ignored.

        The task run() method shall be run within a hooks.task_run()
        context manager to ensure that the task status is recognized by
        the report hooks and other plugins.

        Network executors have additional requirements. See the
        NetworkExecutor class for more information.
        """
        raise NotImplementedError


class LocalExecutor(Executor):
    """
    An Executor that runs a task locally.

    The executor runs the task on the local machine. The task is run
    by calling the task.run() method.

    The executor is created by the local executor factory and is
    typically run sequentially with other executors.
    """

    def __init__(self, factory, task, force_upload=False, force_build=False):
        super().__init__(factory)
        self.task = task
        self.force_build = force_build
        self.force_upload = force_upload

    def schedule(self, env):
        """
        Schedule the task for execution.

        The task is marked as in progress before scheduling.
        """
        self.task.set_in_progress()
        return super().schedule(env)

    def _run(self, env, task):
        if self.is_aborted():
            return
        try:
            with hooks.task_run(task):
                self.task.run(
                    env,
                    force_build=self.force_build,
                    force_upload=self.force_upload)

        except Exception as e:
            log.exception(e, error=False)
            if not task.is_unstable:
                self.task.raise_for_status(log_error=getattr(env, "worker", False))
                raise e

    def get_all_extensions(self, task):
        extensions = copy.copy(task.extensions)
        for ext in extensions:
            extensions.extend(self.get_all_extensions(ext))
        return extensions

    def run(self, env):
        tasks = [self.task] + self.get_all_extensions(self.task)
        for task in tasks:
            task.queued()

        self._run(env, self.task)


class NetworkExecutor(Executor):
    def run(self, env):
        """
        Run the task.

        See the Executor class for basic information.

        Network executors have additional requirements. Before scheduling
        the task to a remote scheduler, the executor must call
        run_resources() on the task. This acquires any Resources marked
        local=True and uploads the resulting session artifacts
        to the remote cache.

        Once the task has been submitted to the remote scheduler, the executor
        must run task.queued() on the task and its extensions. This is done
        to ensure that the task status is correctly reported to the
        user.

        For any change in state of task, the executor must run one of:

        - task.running_execution(remote=True) - when the task has started
        - task.failed_execution(remote=True) - when the task has failed
        - task.failed_execution(remote=True, interrupt=True) - when the
          task has been interrupted, e.g. by a user request or rescheduling
        - task.finished_execution(remote=True) - when the task has passed

        Upon completion of the task, whether successful or not, task
        session artifacts must be downloaded to the local cache, if
        the task is marked as downloadable. This is done by calling
        task.download() with the session_only flag set to True.

        Persistent artifacts are downloaded only if the task is successful
        and the task is marked as downloadable.
        """
        raise NotImplementedError


class SkipTask(Executor):
    """
    An Executor that skips a task.

    This executor is created by the concurrent executor factory when a task
    is skipped, i.e. when the task artifacts are already available locally or
    remotely and the task does not need to be run.
    """

    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def run(self, env):
        """
        Skip the task.

        The task and its extensions are marked as skipped.
        """
        self.task.skipped()
        for ext in self.task.extensions:
            ext.skipped()
        return self.task


class Downloader(Executor):
    """
    An Executor that downloads task artifacts.

    The executor downloads the task artifacts and its extensions from the
    remote cache to the local cache. Failure to download the artifacts
    is reported by raising an exception.

    Downloader executors are typically run in parallel with other executors.

    """
    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def schedule(self, env):
        """
        Schedule the task for execution.

        The task is marked as in progress before scheduling.
        """
        self.task.set_in_progress()
        return super().schedule(env)

    def _download(self, task):
        if self.is_aborted():
            return
        if not task.is_downloadable():
            return
        try:
            task.started_download()
            raise_task_error_if(
                not task.download(persistent_only=True),
                task, "Failed to download task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed_download()
            raise e
        else:
            task.finished_download()

    def run(self, env):
        """ Downloads artifacts. """

        self._download(self.task)
        for ext in self.task.extensions:
            self._download(ext)
        return self.task


class Uploader(Executor):
    """
    An Executor that uploads task artifacts.

    The executor uploads the task artifacts and its extensions from the
    local cache to the remote cache. Failure to upload the artifacts
    is reported by raising an exception.

    Uploader executors are typically run in parallel with other executors.
    """

    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def schedule(self, env):
        """
        Schedule the task for execution.

        The task is marked as in progress before scheduling.
        """
        self.task.set_in_progress()
        return super().schedule(env)

    def _upload(self, task):
        if self.is_aborted():
            return
        try:
            task.started_upload()
            raise_task_error_if(
                not task.upload(persistent_only=True),
                task, "Failed to upload task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed_upload()
            raise e
        else:
            task.finished_upload()

    def run(self, env):
        """ Uploads artifacts. """

        self._upload(self.task)
        for ext in self.task.extensions:
            self._upload(ext)

        return self.task


@utils.Singleton
class ExecutorRegistry(object):
    """
    The ExecutorRegistry is responsible for creating executors.

    The types of executors that are possible to create are:

    - create_local: Runs tasks locally.
    - create_network: Schedules tasks for remote execution.
    - create_downloader: Downloads task artifacts.
    - create_uploader: Uploads task artifacts.
    - create_skipper: Skips tasks.

    The registry utilizes different ExecutorFactory objects to create executors. Plugins
    can register their own NetworkExecutorFactory objects with the help of the
    ExecutorFactory.Register decorator.
    """

    executor_factories = []

    def __init__(self, options=None):
        self._options = options or JoltOptions()
        self._factories = [factory(self._options) for factory in self.__class__.executor_factories]
        self._local_factory = LocalExecutorFactory(self._options)
        self._concurrent_factory = ConcurrentLocalExecutorFactory(self._options)

    def shutdown(self):
        """ Shuts all executor factories and thread-pools down """

        for factory in self._factories:
            factory.shutdown()
        self._local_factory.shutdown()
        self._concurrent_factory.shutdown()

    def create_session(self, graph):
        """ Creates a session for all factories. """
        return {factory: factory.create_session(graph) for factory in self._factories}

    def create_skipper(self, task):
        """ Creates an executor that skips a task. """
        return SkipTask(self._concurrent_factory, task)

    def create_downloader(self, task):
        """ Creates an executor that downloads task artifacts. """
        return Downloader(self._concurrent_factory, task)

    def create_uploader(self, task):
        """ Creates an executor that uploads task artifacts. """
        return Uploader(self._concurrent_factory, task)

    def create_local(self, task, force=False):
        """ Creates an executor that runs a task locally. """
        task.set_locally_executed()
        return self._local_factory.create(task, force=force)

    def create_network(self, session, task):
        """
        Creates an executor that schedules a task for remote execution.

        All registred network executor factories are queried to create an executor.
        The first factory that can create an executor is used. If no factory is able
        to create an executor, a local executor is created as fallback.
        """

        for factory in self._factories:
            executor = factory.create(session[factory], task)
            if executor is not None:
                task.set_remotely_executed()
                return executor
        return self.create_local(task)


class ExecutorFactory(object):
    """
    The ExecutorFactory class is responsible for creating executors.

    The factory is responsible for creating executors that run tasks. The factory
    is also responsible for hosting a thread pool that will run the executors it creates.

    """
    class QueueItem(object):
        """
        The type of item that is put into the queue thread-pool queue.

        It wraps the executor and its priority.
        """
        def __init__(self, priority: int, future: Future, executor: Executor, env: JoltEnvironment):
            self.priority = priority
            self.future = future
            self.executor = executor
            self.env = env

        def __le__(self, o):
            return self.priority <= o.priority

        def __ge__(self, o):
            return self.priority >= o.priority

        def __lt__(self, o):
            return self.priority < o.priority

        def __gt__(self, o):
            return self.priority > o.priority

        def __eq__(self, o):
            return self.priority == o.priority

    @staticmethod
    def Register(cls):
        """
        Decorator to register an executor factory.

        The decorator is used by plugins that whish to register their own
        executor factories. Such factories are used by the ExecutorRegistry
        to create executors for tasks, as determined by the execution strategy
        selected by the user.
        """

        ExecutorRegistry.executor_factories.insert(0, cls)
        return cls

    def __init__(self, options=None, max_workers=None):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self._aborted = False
        self._queue = queue.PriorityQueue()
        self._options = options or JoltOptions()

    def is_aborted(self):
        """ Returns true if the build and thus the factory has been aborted. """
        return self._aborted

    def is_keep_going(self):
        """ Returns true if the build should continue even if a task fails. """
        return self._options.keep_going

    def shutdown(self):
        """
        Called to shutdown the factory and its thread-pool.

        The method is called when the build is complete or when the build is aborted.
        After the method is called, no more tasks can be submitted to the factory and
        the is_aborted() method will return True.
        """
        self._aborted = True
        self.pool.shutdown()

    def create(self, task):
        """
        Create an executor for the provided task.

        Must be implemented by all executor factories. The method must return
        an executor that is capable of running the task. The executor must be
        created with the factory as its parent so that it can be submitted to
        the correct thread-pool for execution.
        """
        raise NotImplementedError()

    def _run(self):
        item = self._queue.get(False)
        self._queue.task_done()
        try:
            if not self.is_aborted():
                item.executor.run(item.env)
        except KeyboardInterrupt as e:
            self._aborted = True
            item.future.set_exception(e)
        except Exception as e:
            if not self.is_keep_going():
                self._aborted = True
            item.future.set_exception(e)
        else:
            item.future.set_result(item.executor)

    def submit(self, executor, env):
        """
        Submit an executor to the thread-pool for execution.

        The method submits the executor to the thread-pool for execution. The executor
        is wrapped in a Future object that is returned to the caller. The Future object
        is used to track the execution of the task and to retrieve the result of the
        execution once it is completed.
        """
        future = Future()
        self._queue.put(ExecutorFactory.QueueItem(-executor.task.weight, future, executor, env))
        self.pool.submit(self._run)
        return future


class LocalExecutorFactory(ExecutorFactory):
    """
    Factory for creating local executors.

    The factory creates executors that run tasks locally. Typically,
    only one LocalExecutor is allowed to run at a time, unless the
    user has specified a higher number of parallel tasks in the
    configuration file or through command line options (-j).
    """

    def __init__(self, options=None):
        max_workers = config.getint(
            "jolt", "parallel_tasks",
            os.getenv("JOLT_PARALLEL_TASKS", 1 if options is None else options.jobs))
        super().__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task, force=False):
        """ Create a LocalExecutor for the task. """
        return LocalExecutor(self, task, force_build=force)


class ConcurrentLocalExecutorFactory(ExecutorFactory):
    """
    A shared factory for local executors that are allowed to run concurrently.

    The factory cannot create any executors on its own. Instead, its executors
    are created by the ExecutorRegistry. The factory thread-pool is then used to
    run executors concurrently.
    """

    def __init__(self, options=None):
        max_workers = tools.Tools().thread_count()
        super().__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task):
        raise NotImplementedError()


class NetworkExecutorFactory(ExecutorFactory):
    """
    Base class for executors that schedule task executions remotely in a build cluster.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(self, session, task):
        raise NotImplementedError()


def ensure_executor_return(func):
    """ Decorator to ensure that an executor is returned by factories. """

    @wraps(func)
    def wrapper(self, session, task):
        executor = func(self, session, task)
        raise_task_error_if(
            not executor, task,
            "no executor can execute the task; "
            "requesting a distributed network build without proper configuration?")
        return executor

    return wrapper


class ExecutionStrategy(object):
    """
    Base class for all execution strategies.

    An execution strategy is responsible for deciding which executor to create for each task.
    The decision is based on the type of task and the availability of the task's artifacts in
    local and remote caches.

    The strategy is also responsible for deciding if task requirements should be pruned
    from the build graph. This is done to avoid processing tasks that are not needed for the build.

    Strategies are selected by the user through command line options.

    """
    def create_executor(self, session, task):
        """
        Create an executor for the task.

        The method must be implemented by all execution strategies. It is responsible for
        creating an executor that is capable of running or processing the task. Creation
        of an executor should be delegated to the ExecutorRegistry which has the knowledge
        of all available executor factories.
        """
        raise NotImplementedError()

    def should_prune_requirements(self, task):
        """
        Return True if the task requirements should be pruned from the build graph.

        The method must be implemented by all execution strategies.
        """
        raise NotImplementedError()


class LocalStrategy(ExecutionStrategy, PruneStrategy):
    """
    Strategy for local builds.

    By default, the strategy schedules tasks for local execution, unless the task
    artifacts are available in the local cache. If available remotely, the strategy
    will create a downloader executor to download the artifacts.
    """

    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    @ensure_executor_return
    def create_executor(self, session, task):
        """ Create an executor for the task. """

        if task.is_alias() or task.is_resource():
            return self.executors.create_skipper(task)
        if not task.is_cacheable():
            return self.executors.create_local(task)
        if task.is_available_locally():
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely():
            return self.executors.create_downloader(task)
        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        """ Prune task requirements if possible """

        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally():
            return True
        if self.cache.download_enabled() and task.is_available_remotely():
            return True
        return False


class DownloadStrategy(ExecutionStrategy, PruneStrategy):
    """
    Strategy for downloading task artifacts.

    The strategy is used when the user has requested that task artifacts be downloaded.
    If the task artifacts are available in the local cache, the strategy will skip the
    task. If the task artifacts are available in the remote cache, the strategy will
    create a downloader executor to download the artifacts. If the task artifacts are
    not available in either cache, the strategy reports an error.
    """

    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    @ensure_executor_return
    def create_executor(self, session, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if task.is_resource():
            return self.executors.create_local(task)
        if not task.is_cacheable():
            return self.executors.create_skipper(task)
        if task.is_available_locally():
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(cache=False):
            return self.executors.create_downloader(task)
        raise_task_error(task, "Task must be built first")

    def should_prune_requirements(self, task):
        return False


class DistributedStrategy(ExecutionStrategy, PruneStrategy):
    """
    Strategy for distributed network builds.

    By default, the strategy schedules tasks for remote execution, if there is no
    artifact available. Otherwise, artifacts are either uploaded or downloaded as
    needed.
    """

    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    @ensure_executor_return
    def create_executor(self, session, task):
        """ Create an executor for the task. """

        if task.is_alias() or task.is_resource():
            return self.executors.create_skipper(task)

        if task.is_local():
            return self.executors.create_local(task)

        if not task.is_cacheable():
            return self.executors.create_network(session, task)

        if not self.cache.upload_enabled():
            return self.executors.create_network(session, task)

        if not task.is_goal(with_extensions=False):
            task.disable_download()
        for extension in task.extensions:
            if not extension.is_goal(with_extensions=False):
                extension.disable_download()

        remote = task.is_available_remotely()
        if remote:
            if task.is_goal() and self.cache.download_enabled() and \
               not task.is_available_locally():
                return self.executors.create_downloader(task)
            return self.executors.create_skipper(task)
        else:
            if task.is_available_locally() and task.is_uploadable():
                return self.executors.create_uploader(task)
            if task.is_fast() and task.deps_available_locally():
                return self.executors.create_local(task, force=True)

        return self.executors.create_network(session, task)

    def should_prune_requirements(self, task):
        """ Prune task requirements if possible """

        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_remotely():
            return True
        return False


class WorkerStrategy(ExecutionStrategy, PruneStrategy):
    """
    Strategy for worker builds.

    This strategy is used on workers when the user has requested a network build.
    It is similar to the LocalStrategy in that it will run tasks locally if no
    artifacts are available. However, if artifacts are available locally, the
    strategy will upload them to the remote cache.
    """

    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    @ensure_executor_return
    def create_executor(self, session, task):
        """ Create an executor for the task. """

        if task.is_alias() or task.is_resource():
            return self.executors.create_skipper(task)

        raise_task_error_if(
            not self.cache.upload_enabled(), task,
            "Artifact upload must be enabled for workers, fix configuration")

        if not task.is_cacheable():
            return self.executors.create_local(task)

        if task.is_available_locally():
            if task.is_goal() and not task.is_available_remotely():
                # Unpacked artifacts may become unpacked before we manage to upload.
                # To keep the implementation simple we take the easy road and rebuild
                # all artifacts that have not been unpacked, even if they are uploadable.
                if task.is_unpacked() and task.is_uploadable():
                    return self.executors.create_uploader(task)
                else:
                    return self.executors.create_local(task, force=True)
            return self.executors.create_skipper(task)

        if not self.cache.download_enabled():
            return self.executors.create_local(task)

        if task.is_available_remotely():
            return self.executors.create_downloader(task)
        elif not task.is_goal():
            raise_task_error(task, "Task artifact removed from global cache, cannot continue")

        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        """ Prune task requirements if possible """

        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally():
            # Unpacked artifacts may become unpacked before we manage to upload.
            # To keep the implementation simple we take the easy road and rebuild
            # all artifacts that have not been unpacked, even if they are uploadable.
            if task.is_unpacked() and task.is_uploadable():
                return True
        if not task.is_goal() and task.task.selfsustained:
            return True
        return False


def get_exported_task_set(task):
    children = [task] + task.descendants
    for ext in task.extensions:
        children.extend(get_exported_task_set(ext))
    return list(set(children))


def export_tasks(tasks):
    pb_tasks = {}

    for task in tasks:
        properties = []
        for key, export in task.task._get_export_objects().items():
            value = export.export(task.task)
            if value is not None:
                pb_attrib = common_pb.Property(key=key, value=str(value))
                properties.append(pb_attrib)

        platform = common_pb.Platform(
            properties=[
                common_pb.Property(key=key, value=value)
                for key, value in task.task.platform.items()
            ]
        )

        args = dict(
            identity=task.identity,
            instance=task.instance,
            taint=str(task.task.taint),
            name=task.short_qualified_name,
            platform=platform,
            properties=properties,
        )

        pb_tasks[task.exported_name] = common_pb.Task(**args)

    return pb_tasks


def export_task_default_params(tasks):
    default_task_names = {}

    for task in tasks:
        for task in task.options.default:
            short_name, params = utils.parse_task_name(task)
            if short_name in default_task_names:
                default_task_names[short_name].update(params)
            else:
                default_task_names[short_name] = params

    return [
        utils.format_task_name(name, params)
        for name, params in default_task_names.items()
    ]



from jolt import attributes, Parameter
from jolt.pkgs import boost
from jolt.plugins import cmake, cxxinfo, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_boost")
@attributes.requires("requires_src")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["mstch"])
class MSTCH(cmake.CMake):
    name = "mstch"
    version = Parameter("1.0.2", help="MSTCH version.")
    requires_boost = ["boost"]
    requires_src = ["fetch:alias=src,url=https://github.com/no1msd/mstch/archive/refs/tags/{version}.tar.gz"]
    srcdir = "{fetch[src]}/mstch-{version}"
    options = [
        "CMAKE_POLICY_VERSION_MINIMUM=3.5",
    ]


TaskRegistry.get().add_task_class(MSTCH)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import cmake
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish(libraries=["ng-log"])
class NgLog(cmake.CMake):
    name = "ng-log"
    version = Parameter("0.8.2", help="ng_log version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    requires_git = ["git:url=https://github.com/ng-log/ng-log.git,rev=v{version}"]
    srcdir = "{git[ng-log]}"
    options = ["BUILD_SHARED_LIBS={shared[ON,OFF]}"]


TaskRegistry.get().add_task_class(NgLog)



from jolt import attributes, Parameter
from jolt.pkgs import cmake, libxml2
from jolt.plugins import cmake, git
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_libxml2")
class Libxslt(cmake.CMake):
    name = "libxslt"
    version = Parameter("1.1.45", help="libxslt version.")
    requires_git = ["git:url=https://gitlab.gnome.org/GNOME/libxslt.git,rev=v{version}"]
    requires_libxml2 = ["libxml2"]
    srcdir = "{git[libxslt]}"


TaskRegistry.get().add_task_class(Libxslt)



from jolt import attributes, Parameter
from jolt.pkgs import cmake, llvm, spirv_tools
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@attributes.requires("requires_llvm")
@attributes.requires("requires_spirv_tools")
@cmake.requires()
@cmake.use_ninja()
class SpirvLLVM(cmake.CMake):
    name = "spirv-llvm-translator"
    version = Parameter("21.1.3", help="SPIRV-LLVM version.")
    requires_git = ["git:url=https://github.com/KhronosGroup/SPIRV-LLVM-Translator.git,rev=v{version}"]
    requires_llvm = ["llvm"]
    requires_spirv_tools = ["spirv-tools"]
    srcdir = "{git[SPIRV-LLVM-Translator]}"


TaskRegistry.get().add_task_class(SpirvLLVM)



from jolt import BooleanParameter, Parameter, attributes
from jolt.plugins import autotools, fetch, pkgconfig
from jolt.tasks import TaskRegistry
import os


@attributes.requires("requires_src")
@autotools.requires()
@pkgconfig.requires()
class Ncurses(autotools.Autotools):
    name = "ncurses"
    version = Parameter("6.6", help="Ncurses version.")
    widechar = BooleanParameter(False, help="Build ncurses with wide character support.")
    shared = BooleanParameter(False, help="Build shared libraries.")
    requires_src = ["fetch:alias=src,url=https://invisible-mirror.net/archives/ncurses/ncurses-{version}.tar.gz"]
    srcdir = "{fetch[src]}/ncurses-{version}"
    options = [
        "--{widechar[enable,disable]}-widec",
        "--{shared[with,without]}-shared",
        "--enable-pc-files",
        "--with-pkg-config-libdir=/jolt-prefix/lib/pkgconfig",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.shared:
            artifact.environ.LD_LIBRARY_PATH.append("lib")


TaskRegistry.get().add_task_class(Ncurses)



from jolt import Task, attributes


@attributes.common_metadata()
class Meson(Task):
    """ Base class for Meson-based build tasks. """

    abstract = True
    """ This is an abstract base class that should be inherited by concrete tasks. """

    incremental = True
    """
    Whether to use incremental builds.
    If True, the build directories are preserved between runs.
    """

    options = []
    """
    Additional options to pass to the `meson` command.
    """

    srcdir = None
    """
    Source directory for the Meson project.

    If None, defaults to the task work directory (joltdir).
    """

    def clean(self, tools):
        at = tools.meson(incremental=self.incremental)
        at.clean()

    def run(self, deps, tools):
        self.deps = deps
        options = tools.expand(self.options)
        at = tools.meson(deps, incremental=self.incremental)
        at.configure(self.srcdir or self.joltdir, *options)
        at.build()
        at.install()

    def publish(self, artifact, tools):
        at = tools.meson(incremental=self.incremental)
        at.publish(artifact)


def requires(meson=True, ninja=True):
    """ Decorator to add Meson and Ninja requirements to a task. """

    import jolt.pkgs.meson
    import jolt.pkgs.ninja

    def decorate(cls):
        if meson:
            cls = attributes.requires("requires_meson")(cls)
            cls.requires_meson = ["meson"]
        if ninja:
            cls = attributes.requires("requires_ninja")(cls)
            cls.requires_ninja = ["ninja"]
        return cls

    return decorate



from jolt import attributes
from jolt.plugins import autotools, git
from jolt.tasks import TaskRegistry


@attributes.common_metadata()
@attributes.requires("requires_src")
@autotools.requires()
class Libffi(autotools.Autotools):
    name = "libffi"
    version = "3.5.2"
    requires_src = ["git:clean=true,url=https://github.com/libffi/libffi.git,rev=v{version}"]
    srcdir = "{git[libffi]}"


TaskRegistry.get().add_task_class(Libffi)



from jolt import attributes, Parameter
from jolt.pkgs import cmake
from jolt.plugins import git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_git")
@cmake.requires()
@cmake.use_ninja()
class FastFloat(cmake.CMake):
    name = "fastfloat"
    version = Parameter("8.1.0", help="fastfloat version.")
    requires_git = ["git:url=https://github.com/fastfloat/fast_float.git,rev=v{version}"]
    srcdir = "{git[fast_float]}"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.cxxinfo.incpaths.append("include")


TaskRegistry.get().add_task_class(FastFloat)



import os
from jolt import attributes, Alias, BooleanParameter, Download, IntParameter, Parameter
from jolt.pkgs import cpython
from jolt.plugins import git
from jolt.tasks import Task, TaskRegistry
from jolt.error import raise_task_error_if


@attributes.common_metadata()
@attributes.requires("requires_git")
@attributes.requires("requires_python_{python[on,off]}")
@attributes.system
class Boost(Task):
    name = "boost"
    version = Parameter("1.90.0", help="Boost version.")
    shared = BooleanParameter(False, help="Build shared libraries")
    python = BooleanParameter(False, help="Build Boost.Python")
    bits = IntParameter(64, values=[32, 64], help="Boost address-model")
    requires_git = ["git:url=https://github.com/boostorg/boost.git,path={buildroot}/git-boost,rev=boost-{version},submodules=true"]
    requires_python_on = ["cpython"]

    def write_user_config(self, deps, tools):
        content = ""

        if self.python:
            py_exe = tools.which("python3")
            raise_task_error_if(
                py_exe is None, self,
                "Boost.Python requested, but no Python interpreter found in PATH.",
            )

            # Find the dependency where the Python interpreter is located
            py_dep = None
            py_prefix = os.path.dirname(os.path.dirname(py_exe))
            for _, dep in deps.items():
                if dep.path == py_prefix:
                    py_dep = dep
                    break

            raise_task_error_if(
                py_dep is None, self,
                "Boost.Python requested, but the Python dependency could not be found.",
            )

            py_version = str(py_dep.strings.version_major)
            content += f"using python : {py_version} : {py_prefix}/bin/python{py_version} : {py_prefix}/include/python{py_version} : {py_prefix}/lib ;\n"

        tools.write_file(
            self.userconfigfile,
            content,
        )

    def clean(self, tools):
        self.builddir = tools.builddir("build", incremental=True)
        self.installdir = tools.builddir("install")
        tools.rmtree(self.builddir)
        tools.rmtree(self.installdir)

    def run(self, deps, tools):
        self.builddir = tools.builddir("build", incremental=True)
        self.installdir = tools.builddir("install")
        self.userconfigfile = tools.expand_path("{builddir}/user-config.jam")
        self.write_user_config(deps, tools)

        without_libs = []
        if not self.python:
            without_libs.append("python")

        bootstrap_cmd = []
        if without_libs:
            bootstrap_cmd = ["--without-libraries={}".format(",".join(without_libs))]

        b2_cmd = [
            "install",
            "address-model={bits}",
            "link={shared[shared,static]}",
            "--prefix={installdir}",
            "--build-dir={builddir}",
            "--user-config={userconfigfile}",
            "-j{}".format(tools.cpu_count()),
        ]

        with tools.cwd("{git[boost]}"):
            if self.system == "windows":
                bootstrap_cmd = [".\\bootstrap.bat", "msvc"] + bootstrap_cmd
                b2_cmd = [".\\b2"] + b2_cmd
            else:
                bootstrap_cmd = ["./bootstrap.sh"] + bootstrap_cmd
                b2_cmd = ["./b2"] + b2_cmd

            tools.run(" ".join(bootstrap_cmd))
            tools.run(" ".join(b2_cmd))

    def publish(self, artifact, tools):
        with tools.cwd(self.installdir):
            artifact.collect("*", symlinks=True)
        artifact.cxxinfo.incpaths.append("include")

        for libdir in ["lib", "lib32", "lib64"]:
            if not os.path.isdir(os.path.join(artifact.path, libdir)):
                continue
            artifact.cxxinfo.libpaths.append(libdir)

            with tools.cwd(self.installdir, libdir):
                for lib in tools.glob("libboost_*.a"):
                    name, _ = os.path.splitext(os.path.basename(lib))
                    artifact.cxxinfo.libraries.append(name[3:])

                arch = tools.getenv("VSCMD_ARG_TGT_ARCH", "x64")
                for lib in tools.glob(f"lib*-mt-{arch}-*.lib"):
                    name, _ = os.path.splitext(lib)
                    artifact.cxxinfo.libraries.append(name)


TaskRegistry.get().add_task_class(Boost)



from jolt import attributes, BooleanParameter, Parameter
from jolt.pkgs import boost, cmake, mysql, sqlite
from jolt.plugins import cxxinfo, git, cmake
from jolt.tasks import TaskRegistry


@attributes.requires("requires_boost_{boost[on,off]}")
@attributes.requires("requires_git")
@attributes.requires("requires_mysql_{mysql[on,off]}")
@attributes.requires("requires_sqlite_{sqlite[on,off]}")
@cmake.requires()
@cmake.use_ninja()
@cxxinfo.publish()
class Soci(cmake.CMake):
    name = "soci"
    version = Parameter("4.1.2", help="Soci version")
    shared = BooleanParameter(False, help="Build shared libraries")
    boost = BooleanParameter(False, help="Enable boost support")
    mysql = BooleanParameter(True, help="Enable MySQL support")
    sqlite = BooleanParameter(True, help="Enable sqlite support")

    requires_boost_on = ["boost"]
    requires_git = ["git:url=https://github.com/SOCI/soci.git,rev=v{version}"]
    requires_mysql_on = ["libmysqlclient"]
    requires_sqlite_on = ["sqlite"]
    srcdir = "{git[soci]}"
    options = [
        "SOCI_TESTS=OFF",
        "SOCI_SHARED={shared[ON,OFF]}",
        "SOCI_SQLITE3={sqlite[ON,OFF]}",
        "SOCI_MYSQL={mysql[ON,OFF]}",
        "WITH_BOOST={boost[ON,OFF]}",
    ]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.sqlite:
            artifact.cxxinfo.libraries.append("soci_sqlite3")
        if self.mysql:
            artifact.cxxinfo.libraries.append("soci_mysql")
        artifact.cxxinfo.libraries.append("soci_odbc")
        artifact.cxxinfo.libraries.append("soci_core")


TaskRegistry.get().add_task_class(Soci)



from jolt import attributes
from jolt.plugins import autotools, fetch
from jolt.tasks import TaskRegistry


@attributes.requires("requires_src")
class Help2Man(autotools.Autotools):
    name = "help2man"
    version = "1.47.17"
    requires_src = ["fetch:alias=src,url=http://ftpmirror.gnu.org/help2man/help2man-{version}.tar.xz"]
    srcdir = "{fetch[src]}/help2man-{version}"


TaskRegistry.get().add_task_class(Help2Man)
